import os
import warnings
from typing import Dict, Tuple

import numpy as np
import torch
from multipledispatch import dispatch
from torch.utils.mobile_optimizer import optimize_for_mobile

from nnsmith.backends.factory import BackendCallable, BackendFactory
from nnsmith.materialize.torch import TorchModel

# Check https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html
# for more PyTorch-internal options.
NNSMITH_PTJIT_OPT_MOBILE = os.getenv("NNSMITH_PTJIT_OPT_MOBILE", "0") == "1"


class TorchJITFactory(BackendFactory):
    def __init__(self, target="cpu", optmax: bool = False, **kwargs):
        super().__init__(target, optmax)
        if self.target == "cpu":
            self.device = torch.device("cpu")
        elif self.target == "cuda":
            self.device = torch.device("cuda")
            torch.backends.cudnn.benchmark = True
        else:
            raise ValueError(
                f"Unknown target: {self.target}. Only `cpu` and `cuda` are supported."
            )

    @property
    def system_name(self) -> str:
        return "torchjit"

    @dispatch(TorchModel)
    def make_backend(self, model: TorchModel) -> BackendCallable:
        torch_net = model.torch_model.to(self.device).eval()
        trace_inp = [ts.to(self.device) for ts in torch_net.get_random_inps().values()]
        with torch.no_grad():
            with warnings.catch_warnings():
                warnings.simplefilter(
                    "ignore",
                    category=torch.jit.TracerWarning,
                )
                exported = torch.jit.trace(torch_net, trace_inp)
                exported = torch.jit.freeze(exported)  # Fronzen graph.
                exported = torch.jit.optimize_for_inference(exported)
                if self.target == "cpu" and NNSMITH_PTJIT_OPT_MOBILE:
                    exported = optimize_for_mobile(exported)

        def closure(inputs: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
            input_ts = {
                k: torch.from_numpy(v).to(self.device) for k, v in inputs.items()
            }
            with torch.no_grad():
                output: Tuple[torch.Tensor] = exported(*input_ts.values())
            return {
                k: v.cpu().detach().resolve_conj().numpy()
                if v.is_conj()
                else v.cpu().detach().numpy()
                for k, v in zip(torch_net.output_like.keys(), output)
            }

        return closure



from functools import reduce
from typing import List, Union

import z3

from nnsmith.abstract.arith import *
from nnsmith.abstract.dtype import DType
from nnsmith.error import ConstraintCheck, SanityCheck


class AbsTensor:
    def __init__(self, shape: List[Union[int, z3.ExprRef]], dtype: DType):
        assert isinstance(
            shape, (list, tuple)
        ), f"Shape must be a list/tuple, but got {shape}"
        self.shape = list(shape)
        self.dtype = DType(dtype)

    def downcast_rank(self):
        return AbsTensor(shape=[None] * self.ndims, dtype=self.dtype)

    def __hash__(self) -> int:
        return hash((tuple(self.shape), self.dtype))

    def __repr__(self) -> str:
        return f"AbsTensor<{self.dtype.short()}>{str(self.shape)}"

    def pretty(self) -> str:
        return f"{self.dtype.short()}{self.shape}"

    def weak_compare(self, other: "AbsTensor") -> bool:
        if self.dtype != other.dtype or self.ndims != other.ndims:
            return False
        for l, r in zip(self.shape, other.shape):
            if isinstance(l, z3.ExprRef) or isinstance(r, z3.ExprRef):
                continue
            if l != r:
                return False
        return True

    def strong_compare(self, other: "AbsTensor") -> bool:
        return self.shape == other.shape and self.dtype == other.dtype

    def __eq__(self, other: "AbsTensor") -> bool:
        return self.strong_compare(other)

    def ge_zero(self):
        ret = []
        for s in self.shape:
            if isinstance(s, z3.ExprRef):
                ret.append(nnsmith_ge(s, 0))
            else:
                ConstraintCheck.ge(s, 0)
        return ret

    def sym_gt_conc_ge_zero(self):
        ret = []
        for s in self.shape:
            if isinstance(s, z3.ExprRef):
                ret.append(nnsmith_gt(s, 0))
            else:
                ConstraintCheck.ge(s, 0)
        return ret

    def gt_zero(self):
        ret = []
        for s in self.shape:
            if isinstance(s, z3.ExprRef):
                ret.append(nnsmith_gt(s, 0))
            else:
                ConstraintCheck.gt(s, 0)
        return ret

    def eq(self, other):
        SanityCheck.eq(self.ndims, other.ndims)
        ret = []
        for i in range(self.ndims):
            if isinstance(self.shape[i], z3.ExprRef) or isinstance(
                other.shape[i], z3.ExprRef
            ):
                ret.append(nnsmith_eq(self.shape[i], other.shape[i]))
            else:
                ConstraintCheck.eq(self.shape[i], other.shape[i])
        return ret

    def torch(self):
        import torch

        return torch.Size(self.shape)

    def constains_symbol(self) -> bool:
        return any(isinstance(s, z3.ExprRef) for s in self.shape)

    def nelement(self):
        if len(self.shape) == 0:  # Scalar
            return 1
        return reduce(lambda x, y: nnsmith_mul(x, y), self.shape, 1)

    def nbytes(self) -> int:
        return self.nelement() * self.dtype.sizeof()

    def deepcopy(self):
        return AbsTensor(shape=list(self.shape), dtype=self.dtype)

    @property
    def ndims(self):
        return len(self.shape)

    def is_concrete(self) -> bool:
        return all(isinstance(s, int) for s in self.shape)

    def htype(self):  # High-level type
        return (self.dtype, self.ndims)



from typing import Union

import z3

from nnsmith.error import ConstraintCheck, SanityCheck

ARITH_MAX_WIDTH: int = 64


def align_bvs(
    left: Union[float, int, z3.ExprRef],
    right: Union[float, int, z3.ExprRef],
    carry=False,
    mult=False,
):
    left_is_arith = isinstance(left, (int, float, z3.ArithRef))
    right_is_arith = isinstance(right, (int, float, z3.ArithRef))
    # If both values are of arithmetic type, we do not need to do anything.
    if left_is_arith and right_is_arith:
        return (left, right)
    # We assume that the width of an arithmetic type is ARITH_MAX_WIDTH.
    if left_is_arith:
        if isinstance(left, int):
            left_size = min(ARITH_MAX_WIDTH, left.bit_length())
        else:
            left_size = ARITH_MAX_WIDTH
    elif isinstance(left, z3.BitVecRef):
        left_size = left.size()
    else:
        raise RuntimeError(f"Unsupported alignment value {left} of type {type(left)}")
    # We assume that the width of an arithmetic type is ARITH_MAX_WIDTH.
    if right_is_arith:
        if isinstance(right, int):
            right_size = min(ARITH_MAX_WIDTH, right.bit_length())
        else:
            right_size = ARITH_MAX_WIDTH
    elif isinstance(right, z3.BitVecRef):
        right_size = right.size()
    else:
        raise RuntimeError(f"Unsupported alignment value {right} of type {type(right)}")
    # Extend the bitvector that is smaller with the necessary amount of zeroes.
    SanityCheck.true(
        not (carry and mult),
        "Carry and multiplication extension are mutually exclusive",
    )
    SanityCheck.le(
        left_size,
        ARITH_MAX_WIDTH,
        f"Bitvector sizes must not exceed {ARITH_MAX_WIDTH} bits.",
    )
    SanityCheck.le(
        right_size,
        ARITH_MAX_WIDTH,
        f"Bitvector sizes must not exceed {ARITH_MAX_WIDTH} bits.",
    )
    diff = left_size - right_size
    if left_is_arith:
        if diff > 0:
            right = z3.Concat(z3.BitVecVal(0, diff), right)
        if isinstance(left, z3.IntNumRef):
            left = left.as_long()
        return z3.BitVecVal(left, right.size()), z3.simplify(right)
    if right_is_arith:
        if diff < 0:
            left = z3.Concat(z3.BitVecVal(0, abs(diff)), left)
        if isinstance(left, z3.IntNumRef):
            left = left.as_long()
        return left, z3.BitVecVal(right, left.size())
    if diff < 0:
        left = z3.Concat(z3.BitVecVal(0, abs(diff)), left)
    elif diff > 0:
        right = z3.Concat(z3.BitVecVal(0, diff), right)

    if carry and max(left_size, right_size) < ARITH_MAX_WIDTH:
        left = z3.Concat(z3.BitVecVal(0, 1), left)
        right = z3.Concat(z3.BitVecVal(0, 1), right)
    if mult:
        max_val = right.size() + left.size()
        if max_val >= ARITH_MAX_WIDTH:
            return (left, right)
        else:
            max_val = ARITH_MAX_WIDTH - max_val
        left = z3.Concat(z3.BitVecVal(0, max_val), left)
        right = z3.Concat(z3.BitVecVal(0, max_val), right)
    return (left, right)


def nnsmith_mul(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right, mult=True)
    return left * right


def nnsmith_add(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right, carry=True)
    return left + right


def nnsmith_sub(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    return left - right


def nnsmith_eq(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    return left == right


def nnsmith_neq(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    return left != right


def nnsmith_ge(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    if isinstance(left, z3.BitVecRef) or isinstance(right, z3.BitVecRef):
        return z3.UGE(left, right)
    return left >= right


def nnsmith_gt(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    if isinstance(left, z3.BitVecRef) or isinstance(right, z3.BitVecRef):
        return z3.UGT(left, right)
    return left > right


def nnsmith_le(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    if isinstance(left, z3.BitVecRef) or isinstance(right, z3.BitVecRef):
        return z3.ULE(left, right)
    return left <= right


def nnsmith_lt(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    if isinstance(left, z3.BitVecRef) or isinstance(right, z3.BitVecRef):
        return z3.ULT(left, right)
    return left < right


def nnsmith_div(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    if isinstance(left, z3.BitVecRef) or isinstance(right, z3.BitVecRef):
        return z3.UDiv(left, right)
    if isinstance(left, int) and isinstance(right, int):
        ConstraintCheck.true(right != 0, "Div by zero")
        return left // right
    return left / right


def nnsmith_mod(
    left: Union[float, int, z3.ExprRef], right: Union[float, int, z3.ExprRef]
):
    left, right = align_bvs(left, right)
    if isinstance(left, z3.BitVecRef) or isinstance(right, z3.BitVecRef):
        return z3.URem(left, right)
    if isinstance(right, int):
        ConstraintCheck.true(right != 0, "Mod by zero")
    return left % right


def nnsmith_min(left, right):
    if isinstance(left, int) and isinstance(right, int):
        return min(left, right)
    left, right = align_bvs(left, right)
    return z3.If(nnsmith_le(left, right), left, right)


def nnsmith_max(left, right):
    if isinstance(left, int) and isinstance(right, int):
        return max(left, right)
    left, right = align_bvs(left, right)
    return z3.If(nnsmith_ge(left, right), left, right)


def nnsmith_and(left, right):
    if isinstance(left, bool) and isinstance(right, bool):
        return left and right
    return z3.And(left, right)


def nnsmith_or(left, right):
    if isinstance(left, bool) and isinstance(right, bool):
        return left or right
    return z3.Or(left, right)


def nnsmith_not(expr):
    if isinstance(expr, bool):
        return not expr
    return z3.Not(expr)



"""Evaluating input searching algorithms.
simply generate a bunch of models and see if the can find viable inputs.
"""

from multiprocessing import Process
import shutil
import traceback
import uuid
import time
import argparse
import random
import os
import pickle
import psutil

import torch
import numpy as np
import pandas as pd
from tqdm import tqdm
import networkx as nx


from nnsmith.materialize.onnx.export import torch2onnx
from nnsmith.util import mkdir
from nnsmith.graph_gen import random_model_gen, SymbolNet, random_tensor
from nnsmith.materialize.torch.input_gen import GradSearch, SamplingSearch
from nnsmith.dtype_test import rewrite_op_dtype
from nnsmith.abstract.op import ALL_OP_TYPES


def mknet(args, differentiable_ops=None):
    model_seed = random.getrandbits(32)
    gen, solution = random_model_gen(
        mode=args.mode,
        seed=model_seed,
        max_nodes=args.max_nodes,
        candidates_overwrite=differentiable_ops,
        init_fp=True,
        skip=args.skip,
    )
    net = SymbolNet(
        gen.abstract_graph,
        solution,
        verbose=args.verbose,
        alive_shapes=gen.alive_shapes,
    )
    net.eval()
    return net, gen.num_op(), model_seed


def mknets(args, exp_seed):
    np.random.seed(exp_seed)
    random.seed(exp_seed)
    torch.manual_seed(exp_seed)

    model_path = os.path.join(args.root, "model")
    if not os.path.exists(model_path):
        os.makedirs(model_path)
    __DIFF_CACHE__ = "config/diff.pkl"
    differentiable_ops = rewrite_op_dtype(
        ALL_OP_TYPES, factory=None, diff=True, verbose=True, cache=__DIFF_CACHE__
    )
    print(differentiable_ops)
    results = {
        "model_seed": [],
        "n_nodes": [],
        "gen_time": [],
    }
    for model_id in tqdm(range(args.n_model)):
        gen_time = None
        while True:
            try:
                gen_tstart = time.time()
                net, num_op, model_seed = mknet(args, differentiable_ops)
            except:
                traceback.print_exc()
                continue
            # break # NOTE: uncomment this line to see how serious the issue is.
            if net.n_vulnerable_op > 0:
                try:
                    torch2onnx(net, os.path.join(model_path, f"{model_id}.onnx"))
                except Exception as e:
                    print(e)
                    print("Failed to convert to onnx")
                gen_time = time.time() - gen_tstart
                break
        if hasattr(net, "graph"):
            nx.drawing.nx_pydot.to_pydot(net.graph).write_png(
                os.path.join(model_path, f"{model_id}-graph.png")
            )
            net.to_picklable()
        cloudpickle.dump(
            net, open(os.path.join(model_path, f"{model_id}-net.pkl"), "wb"), protocol=4
        )
        results["n_nodes"].append(num_op)
        results["model_seed"].append(model_seed)
        results["gen_time"].append(gen_time)
    df = pd.DataFrame(results)
    df.to_csv(os.path.join(args.root, "model_info.csv"), index=False)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--max_nodes", type=int, default=20)
    parser.add_argument("--exp-seed", type=int)
    parser.add_argument("--max_sample", type=int, default=1)
    parser.add_argument("--max_time_ms", type=int, default=500)
    parser.add_argument("--n_model", type=int, default=50)
    parser.add_argument("--use_cuda", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("--mode", type=str, default="random")
    parser.add_argument("--root", type=str, help="save models and results to this path")
    parser.add_argument("--skip")
    # for reproducibility
    parser.add_argument(
        "--result", help="Name of the grad search result csv (exp_name).", default=None
    )
    parser.add_argument("--print_grad", type=int, default=0)
    args = parser.parse_args()

    MAX_MEM = (
        psutil.virtual_memory().total
        if not args.use_cuda
        else torch.cuda.get_device_properties(0).total_memory
    )

    del_root = False
    if args.root is None:
        args.root = "input_search_root_" + str(uuid.uuid4())
        del_root = True

    exp_seed = args.exp_seed
    if exp_seed is None:
        exp_seed = random.getrandbits(32)
    print(f"Using seed {exp_seed}")

    # mkdir and generate models if not exists.
    if not os.path.exists(args.root):
        mkdir(args.root)
        p = Process(target=mknets, args=(args, exp_seed))
        p.start()
        p.join()
    else:
        print(f"{args.root} exists, skip model generation phase.")

    np.random.seed(exp_seed)
    random.seed(exp_seed)
    torch.manual_seed(exp_seed)

    ref_df = pd.read_csv(os.path.join(args.root, "model_info.csv"))  # load models

    exp_name = (
        args.result
        if args.result
        else f"r{exp_seed}-model{args.n_model}-node{args.max_nodes}-inp-search.csv"
    )
    if not exp_name.endswith(".csv"):
        exp_name += ".csv"

    results = {
        "model_seed": [],
        "n_nodes": [],
        "naive-time": [],
        "naive-try": [],
        "naive-succ": [],
        "sampling-time": [],
        "sampling-try": [],
        "sampling-succ": [],
        "grad-time": [],
        "grad-try": [],
        "grad-succ": [],
        "proxy-time": [],
        "proxy-try": [],
        "proxy-succ": [],
    }

    with tqdm(range(args.n_model)) as pbar:
        for model_id in pbar:
            model_seed = ref_df["model_seed"][model_id]
            pbar.set_description(f"id={model_id}, seed={model_seed}")

            net = pickle.load(
                open(os.path.join(args.root, "model", f"{model_id}-net.pkl"), "rb")
            )
            net = SymbolNet(
                net.concrete_graph,
                None,
                verbose=args.verbose,
                megabyte_lim=net.megabyte_lim,
                print_grad=args.print_grad,
            )
            net.eval()
            net.use_gradient = False
            num_op = ref_df["n_nodes"][model_id]

            results["n_nodes"].append(num_op)
            results["model_seed"].append(model_seed)

            def seedme():
                nseed = (model_seed + 1) % (2**32)
                random.seed(nseed)
                np.random.seed(nseed)
                torch.manual_seed(nseed)

            # ------------------------------------------------------------
            # Test naive:
            # how other fuzzers do: just randomly init tensors using torch.random -> 0~1 by `once`
            seedme()
            strt_time = time.time()

            if args.use_cuda:
                net.use_cuda()

            # ------------------------------------------------------------
            # Estimate model size | avoid OOM
            nbytes = 0
            for ii in net.input_info:
                dtype = ii.op.asb_tensor.dtype.torch()
                nbytes += (
                    ii.op.asb_tensor.nelement()
                    * torch.tensor(0, dtype=dtype).element_size()
                )
            for name, param in net.named_parameters():
                nbytes += param.numel() * param.element_size()

            MEM_FACTOR = 0.4
            if nbytes * args.max_sample > MAX_MEM * MEM_FACTOR:
                max_sample = int(MAX_MEM * MEM_FACTOR / nbytes)
                print(
                    f"{args.max_sample} x weights require {nbytes * args.max_sample / (1024**3)}GB"
                    f" which is more than what you have. Downgrade to {max_sample} samples."
                )
            else:
                max_sample = args.max_sample
            # ------------------------------------------------------------
            seedme()

            net.check_intermediate_numeric = True
            with torch.no_grad():
                _ = net(
                    *net.get_random_inps(
                        base="center", margin=1, use_cuda=args.use_cuda
                    )
                )
                results["naive-succ"].append(not net.invalid_found_last)

            results["naive-time"].append(time.time() - strt_time)
            results["naive-try"].append(1)  # naive method always try once
            # ------------------------------------------------------------

            # ------------------------------------------------------------
            # Test sampling:
            # sync input & weight between `sampling` and `grad`
            seedme()
            init_tensor_samples = []
            for i in range(max_sample):
                data = net.get_random_inps(use_cuda=args.use_cuda)
                init_tensor_samples.append(data)

            init_weight_samples = []
            with torch.no_grad():
                for i in range(max_sample):
                    weight_sample = {}
                    for name, param in net.named_parameters():
                        weight_sample[name] = random_tensor(
                            param.shape, dtype=param.dtype, use_cuda=args.use_cuda
                        )
                    init_weight_samples.append(weight_sample)

            def apply_weights(net, weight_sample):
                with torch.no_grad():
                    for name, param in net.named_parameters():
                        param.copy_(weight_sample[name])

            if args.use_cuda:
                net.use_cuda()

            strt_time = time.time()
            searcher = SamplingSearch(
                net, init_tensor_samples, init_weight_samples, use_cuda=args.use_cuda
            )

            n_try, sat_inputs = searcher.search(
                max_time_ms=args.max_time_ms, max_sample=args.max_sample
            )

            results["sampling-time"].append(time.time() - strt_time)
            results["sampling-try"].append(n_try)
            results["sampling-succ"].append(sat_inputs is not None)
            # ------------------------------------------------------------

            # ------------------------------------------------------------
            # Test grad
            # If sampling can succeed, grad can succeed too as their initial input are the same.
            seedme()

            strt_time = time.time()

            searcher = GradSearch(
                net, init_tensor_samples, init_weight_samples, use_cuda=args.use_cuda
            )
            n_try, sat_inputs = searcher.search(
                max_time_ms=args.max_time_ms, max_sample=args.max_sample
            )

            results["grad-time"].append(time.time() - strt_time)
            results["grad-try"].append(n_try)
            results["grad-succ"].append(sat_inputs is not None)
            # --------------------------------------------------------------------

            # ------------------------------------------------------------
            # Test grad + proxy
            # If sampling can succeed, grad can succeed too as their initial input are the same.
            # Proxy makes some operators differentiable.
            seedme()

            strt_time = time.time()

            net.enable_proxy_grad()
            searcher = GradSearch(
                net, init_tensor_samples, init_weight_samples, use_cuda=args.use_cuda
            )
            n_try, sat_inputs = searcher.search(
                max_time_ms=args.max_time_ms, max_sample=args.max_sample
            )

            results["proxy-time"].append(time.time() - strt_time)
            results["proxy-try"].append(n_try)
            results["proxy-succ"].append(sat_inputs is not None)
            # --------------------------------------------------------------------

    df = pd.DataFrame(results)
    df.to_csv(os.path.join(args.root, exp_name), index=False)
    with open(os.path.join(args.root, exp_name + "-stats.log"), "w") as f:
        f.write(str(df.mean()) + "\n")
    with pd.option_context("display.float_format", "{:0.3f}".format):
        print(df.drop("model_seed", axis=1).mean())
    if del_root:
        shutil.rmtree(args.root)



import logging
from abc import ABC, abstractmethod


class InternalError(Exception):
    """Fatal unexpected internal errors in NNSmith that should shut down the program immediately."""

    pass


class ConstraintError(Exception):
    """Expected possible constarint unsat error used in shape transfer function."""

    pass


class BaseChecker(ABC):
    @classmethod
    @abstractmethod
    def handler(cls, msg):
        pass

    @classmethod
    def eq(cls, lhs, rhs, msg=""):
        if lhs != rhs:
            cls.handler(f"Failed asertion :: {msg} | {lhs} != {rhs}")

    @classmethod
    def gt(cls, lhs, rhs, msg=""):
        if lhs <= rhs:
            cls.handler(f"Failed asertion :: {msg} | {lhs} <= {rhs}")

    @classmethod
    def ge(cls, lhs, rhs, msg=""):
        if lhs < rhs:
            cls.handler(f"Failed asertion :: {msg} | {lhs} < {rhs}")

    @classmethod
    def lt(cls, lhs, rhs, msg=""):
        if lhs >= rhs:
            cls.handler(f"Failed asertion :: {msg} | {lhs} >= {rhs}")

    @classmethod
    def le(cls, lhs, rhs, msg=""):
        if lhs > rhs:
            cls.handler(f"Failed asertion :: {msg} | {lhs} > {rhs}")

    @classmethod
    def none(cls, obj, msg=""):
        if obj is not None:
            cls.handler(f"Failed asertion :: {msg} | expr is not None")

    @classmethod
    def not_none(cls, obj, msg=""):
        if obj is None:
            cls.handler(f"Failed asertion :: {msg} | expr is None")

    @classmethod
    def true(cls, cond, msg=""):
        if not cond:
            cls.handler(f"Failed asertion :: {msg} | condition is not True")

    @classmethod
    def false(cls, cond, msg=""):
        if cond:
            cls.handler(f"Failed asertion :: {msg} | condition is not False")


class SanityCheck(BaseChecker):
    @classmethod
    def handler(cls, msg):
        logging.critical(msg)
        raise InternalError(
            msg + " | Reporting bugs @ https://github.com/ise-uiuc/nnsmith/issues"
        )


class ConstraintCheck(BaseChecker):
    @classmethod
    def handler(cls, msg):
        raise ConstraintError(msg)



from functools import partial
from typing import Type

import torch

from nnsmith.abstract.dtype import DTYPE_GEN_INTS
from nnsmith.abstract.op import *
from nnsmith.materialize import framework_operator_impl
from nnsmith.materialize.torch.dialect import Flatten, Linear, TorchReduceSum

# Implementation of operators.

# core dialect + some future PyTorch-only Operators.
TORCH_REALIZABLE_OPS = FULL_OPERATOR_SETS["core"] + FULL_OPERATOR_SETS["torch"]
ALL_TORCH_OPS: List[Type[AbsOpBase]] = []

operator_impl = partial(framework_operator_impl, TORCH_REALIZABLE_OPS, ALL_TORCH_OPS)

# forward_fn:  forward


@operator_impl(Constant)
def forward_fn(op: Constant):
    class ConstFn(torch.nn.Module):
        def __init__(self, data: torch.Tensor) -> None:
            super().__init__()
            self.data = torch.nn.parameter.Parameter(
                data, requires_grad=data.is_floating_point()
            )

        def forward(self):
            return self.data

    return ConstFn(torch.randn(op.abs_tensor.shape).to(op.abs_tensor.dtype.torch()))


@operator_impl(ReLU)
def forward_fn(op: ReLU):
    return torch.relu


@operator_impl(GELU)
def forward_fn(op: GELU):
    return torch.nn.functional.gelu


@operator_impl(LeakyReLU)
def forward_fn(op: LeakyReLU):
    return torch.nn.LeakyReLU(0.01)


@operator_impl(PReLU)
def forward_fn(op: PReLU):
    return torch.nn.PReLU()


@operator_impl(Sigmoid)
def forward_fn(op: Sigmoid):
    return torch.sigmoid


@operator_impl(Sin)
def forward_fn(op: Sin):
    return torch.sin


@operator_impl(Cos)
def forward_fn(op: Cos):
    return torch.cos


@operator_impl(Asin)
def forward_fn(op: Asin):
    return torch.asin


@operator_impl(Acos)
def forward_fn(op: Acos):
    return torch.acos


@operator_impl(Tan)
def forward_fn(op: Tan):
    return torch.tan


@operator_impl(Atan)
def forward_fn(op: Atan):
    return torch.atan


# Abs
@operator_impl(Abs)
def forward_fn(op: Abs):
    return torch.abs


@operator_impl(Where)
def forward_fn(op: Where):
    return torch.where


@operator_impl(Add)
def forward_fn(op: Add):
    return torch.add


@operator_impl(Sub)
def forward_fn(op: Sub):
    return torch.sub


@operator_impl(Mul)
def forward_fn(op: Mul):
    return torch.mul


@operator_impl(Div)
def forward_fn(op: Div):
    if op.input_like[0].dtype in DTYPE_GEN_INTS:
        return lambda up, down: torch.div(up, down, rounding_mode="floor")
    return torch.div


@operator_impl(Max)
def forward_fn(op: Max):
    return torch.max


@operator_impl(Min)
def forward_fn(op: Min):
    return torch.min


@operator_impl(Equal)
def forward_fn(op: Equal):
    return torch.eq


@operator_impl(Greater)
def forward_fn(op: Greater):
    return torch.gt


@operator_impl(Less)
def forward_fn(op: Less):
    return torch.lt


@operator_impl(And)
def forward_fn(op: And):
    return torch.logical_and


@operator_impl(Or)
def forward_fn(op: Or):
    return torch.logical_or


@operator_impl(Xor)
def forward_fn(op: Xor):
    return torch.logical_xor


@operator_impl(Pow)
def forward_fn(op: Pow):
    return torch.pow


# Floor
@operator_impl(Floor)
def forward_fn(op: Floor):
    return torch.floor


# Ceil
@operator_impl(Ceil)
def forward_fn(op: Ceil):
    return torch.ceil


@operator_impl(Clip)
def forward_fn(op: Clip):
    if op.input_like[0].dtype in DTYPE_GEN_FLOATS:
        return lambda x: torch.clip(x, -1.5, 1.5)
    else:
        return lambda x: torch.clip(x, -1, 1)


@operator_impl(Round)
def forward_fn(op: Round):
    return torch.round


@operator_impl(Sqrt)
def forward_fn(op: Sqrt):
    return torch.sqrt


@operator_impl(Log2)
def forward_fn(op: Log2):
    return torch.log2


@operator_impl(Neg)
def forward_fn(op: Neg):
    return torch.neg


@operator_impl(Softmax)
def forward_fn(op: Softmax):
    return torch.nn.Softmax(dim=op.dim)


@operator_impl(MaxPool2d)
def forward_fn(op: MaxPool2d):
    return torch.nn.MaxPool2d(
        kernel_size=(op.kernel_h_size, op.kernel_w_size),
        stride=op.stride,
        padding=op.padding,
    )


@operator_impl(AvgPool2d)
def forward_fn(op: AvgPool2d):
    return torch.nn.AvgPool2d(
        kernel_size=(op.kernel_h_size, op.kernel_w_size),
        stride=op.stride,
        padding=op.padding,
    )


@operator_impl(Slice)
def forward_fn(op: Slice):
    reg = op.extra_attrs["region"]
    shape = op.input_like[0].shape
    dim_s = shape[op.extra_attrs["axis"]]
    start, end = op.start, op.end
    if reg in ["left", "mid"]:
        start -= dim_s
    # actual end would be 0, which is not really 'left'
    if reg == "left" and end < dim_s and end != Slice.INT_MAX:
        end -= dim_s
    idx = tuple(
        slice(None, None) if i != op.extra_attrs["axis"] else slice(start, end, op.step)
        for i in range(op.extra_attrs["ndims"])
    )

    return lambda x: x[idx]


@operator_impl(Pad)
def forward_fn(op: Pad):
    if op.extra_attrs["type"] == "constant":
        # 0 easily cause division by zero...
        # 1 easily cause false positives (sqrt(1) = 0.99999... != 1 in ORT, so floor(sqrt(1))=0)
        return lambda x: torch.nn.functional.pad(
            x, op.padding_list, "constant", value=0.5
        )
    elif op.extra_attrs["type"] == "replicate" or op.extra_attrs["type"] == "reflect":
        return lambda x: torch.nn.functional.pad(
            x, op.padding_list, op.extra_attrs["type"]
        )


@operator_impl(Expand)
def forward_fn(op: Expand):
    return lambda x: x.expand(*op.output_like[0].shape)


@operator_impl(BatchNorm2d)
def forward_fn(op: BatchNorm2d):
    return torch.nn.BatchNorm2d(num_features=op.nfeat)


@operator_impl(Conv1d)
def forward_fn(op: Conv1d):
    return torch.nn.Conv1d(
        in_channels=op.in_channels,
        out_channels=op.out_channels,
        kernel_size=op.kernel_size,
        stride=op.stride,
        padding=op.padding,
        dilation=op.dilation,
    )


@operator_impl(NCHWConv2d)
def forward_fn(op: NCHWConv2d):
    return torch.nn.Conv2d(
        op.in_channels,
        op.out_channels,
        kernel_size=(op.kernel_h_size, op.kernel_w_size),
        stride=op.stride,
        padding=op.padding,
        dilation=(op.dilation_h, op.dilation_w),
    )


@operator_impl(Reshape)
def forward_fn(op: Reshape):
    return lambda x: x.reshape(*op.target_shape)


@operator_impl(Flatten)
def forward_fn(op: Flatten):
    return torch.Tensor.flatten


@operator_impl(Transpose)
def forward_fn(op: Transpose):
    dim0, dim1 = op._init_swap_dims(op.input_like[0].shape)
    return lambda x: x.transpose(dim0, dim1)


# NearestInterp
@operator_impl(NearestInterp)
def forward_fn(op: NearestInterp):
    return lambda x: torch.nn.functional.interpolate(x, size=op.size, mode="nearest")


# LinearInterp
@operator_impl(LinearInterp)
def forward_fn(op: LinearInterp):
    return lambda x: torch.nn.functional.interpolate(x, size=op.size, mode="linear")


# BilinearInterp
@operator_impl(BilinearInterp)
def forward_fn(op: BilinearInterp):
    return lambda x: torch.nn.functional.interpolate(x, size=op.size, mode="bilinear")


@operator_impl(BicubicInterp)
def forward_fn(op: BicubicInterp):
    return lambda x: torch.nn.functional.interpolate(x, size=op.size, mode="bicubic")


# TrilinearInterp
@operator_impl(TrilinearInterp)
def forward_fn(op: TrilinearInterp):
    return lambda x: torch.nn.functional.interpolate(x, size=op.size, mode="trilinear")


@operator_impl(Squeeze)
def forward_fn(op: Squeeze):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: x.squeeze(op.extra_attrs["reduce_dim"])
    return torch.Tensor.squeeze


@operator_impl(TorchReduceSum)
def forward_fn(op: TorchReduceSum):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: x.sum(op.extra_attrs["reduce_dim"])
    return torch.Tensor.sum


# ReduceMin
@operator_impl(ReduceMin)
def forward_fn(op: ReduceMin):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: x.min(op.extra_attrs["reduce_dim"]).values
    return torch.Tensor.min


# ReduceMax
@operator_impl(ReduceMax)
def forward_fn(op: ReduceMax):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: x.max(op.extra_attrs["reduce_dim"]).values
    return torch.Tensor.max


# ReduceMean
@operator_impl(ReduceMean)
def forward_fn(op: ReduceMean):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: x.mean(op.extra_attrs["reduce_dim"])
    return torch.Tensor.mean


# ArgMin
@operator_impl(ArgMin)
def forward_fn(op: ArgMin):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: x.argmin(op.extra_attrs["reduce_dim"])
    return torch.Tensor.argmin


# ArgMax
@operator_impl(ArgMax)
def forward_fn(op: ArgMax):
    if op.extra_attrs["reduce_dim"] is not None:
        return lambda x: x.argmax(op.extra_attrs["reduce_dim"])
    return torch.Tensor.argmax


# Tril
@operator_impl(Tril)
def forward_fn(op: Tril):
    return lambda x: x.tril(op.diagonal)


# Triu
@operator_impl(Triu)
def forward_fn(op: Triu):
    return lambda x: x.triu(op.diagonal)


# Linear
@operator_impl(Linear)
def forward_fn(op: Linear):
    return torch.nn.Linear(in_features=op.ifeat, out_features=op.ofeat)


@operator_impl(Concat)
def forward_fn(op: Concat):
    axis = op.extra_attrs["axis"]
    return lambda *args: torch.cat(args, dim=axis)


@operator_impl(Cast)
def forward_fn(op: Cast):
    return lambda x: x.to(dtype=op.extra_attrs["to"].torch())


@operator_impl(MatMul)
def forward_fn(op: MatMul):
    return torch.matmul



"""Directly perform model execution from a bare model, NNSmith model folder, or NNSmith bug report folder.
Example usage:
# compile + run the model (in ORT by default); verify with the oracle under `nnsmith_output`;
 nnsmith.model_exec model.path=nnsmith_output/model.onnx
# compile + run the model; verify with the oracle with fallback mode (random input)
 nnsmith.model_exec model.path=nnsmith_output/model.onnx cmp.oracle=null
# differential testing with tvm
 nnsmith.model_exec model.path=nnsmith_output/model.onnx cmp.with='{type:tvm, optmax:true, target:cpu}'
"""

import os
import pickle
import random
from pathlib import Path
from types import FunctionType
from typing import List, Optional

import hydra
from omegaconf import DictConfig, ListConfig

from nnsmith.backends import BackendFactory
from nnsmith.logging import EXEC_LOG
from nnsmith.macro import NNSMITH_BUG_PATTERN_TOKEN
from nnsmith.materialize import BugReport, Model, Oracle, TestCase


def verify_testcase(
    cmp_cfg: DictConfig,
    factory: BackendFactory,
    testcase: TestCase,
    output_dir: os.PathLike,
    filters: List = None,
    supress_succ=True,
    crash_safe: bool = False,
    timeout: Optional[int] = None,
) -> bool:
    if filters is None:
        filters = []

    def check_result(bug_report_or, odir, msg=None) -> bool:  # succ?
        msg = "" if msg is None else msg
        if not isinstance(bug_report_or, BugReport):
            if not supress_succ:
                EXEC_LOG.info(f"[PASS] {msg}")
            return True
        else:
            bug_report = bug_report_or

            for f in filters:
                if f(bug_report):  # filter: no log & dump. but still a bug.
                    filter_name = (
                        f.__name__ if isinstance(f, FunctionType) else type(f).__name__
                    )
                    EXEC_LOG.info(
                        f"Filter [{filter_name}] {bug_report.symptom} at {bug_report.stage}"
                    )
                    return False

            EXEC_LOG.warning("[FAIL] ")
            EXEC_LOG.warning(bug_report.log)
            if odir is not None:
                odir = str(odir).replace(
                    NNSMITH_BUG_PATTERN_TOKEN,
                    f"{bug_report.symptom}-{bug_report.stage}",
                )
                EXEC_LOG.warning("Saving bug report to {}".format(odir))
                bug_report.dump(odir)
            return False

    bug_or_res = factory.checked_compile_and_exec(
        testcase, crash_safe=crash_safe, timeout=timeout
    )
    if check_result(
        bug_or_res, odir=output_dir, msg=f"Compile + Execution [{factory}]"
    ):
        if testcase.oracle.output is not None:  # we have output results && no bug yet
            # do result verification
            if not check_result(
                factory.verify_results(
                    bug_or_res,
                    testcase,
                    equal_nan=cmp_cfg["equal_nan"],
                ),
                odir=output_dir,
                msg=f"Result Verification w/ Oracle from {testcase.oracle.provider}",
            ):
                return False
    else:
        return False

    # Compare with
    if cmp_cfg["with"] is not None and cmp_cfg["with"]["type"] is not None:
        cmp_fac = BackendFactory.init(
            cmp_cfg["with"]["type"],
            target=cmp_cfg["with"]["target"],
            optmax=cmp_cfg["with"]["optmax"],
        )
        cmp_testcase = cmp_fac.make_testcase(
            testcase.model,
            input=testcase.oracle.input,
            crash_safe=crash_safe,
            timeout=timeout,
        )
        if check_result(
            cmp_testcase,
            odir=output_dir,
            msg=f"Compile + Execution [`cmp.with`: {cmp_fac}]",
        ):
            if not check_result(
                factory.verify_results(
                    bug_or_res,
                    cmp_testcase,
                    equal_nan=cmp_cfg["equal_nan"],
                ),
                odir=output_dir,
                msg="Result Verification w/ Reference Backend",
            ):
                return False
        else:
            return False

    return True


@hydra.main(version_base=None, config_path="../config", config_name="main")
def main(cfg: DictConfig):
    cmp_cfg = cfg["cmp"]
    seed = random.getrandbits(32) if cmp_cfg["seed"] is None else cmp_cfg["seed"]
    EXEC_LOG.info(f"Using seed {seed}")

    model_cfg = cfg["model"]
    ModelType = Model.init(model_cfg["type"], cfg["backend"]["target"])
    ModelType.add_seed_setter()

    if isinstance(model_cfg["path"], ListConfig):
        model_paths = model_cfg["path"]
        if cmp_cfg["save"] is not None:
            assert isinstance(cmp_cfg["save"], ListConfig), (
                "With multiple models compiled-and-executed together,"
                " the `save` must also be a list of paths."
            )
    else:
        model_paths = [model_cfg["path"]]

    output_dirs = cmp_cfg["save"]
    if isinstance(output_dirs, (int, float, str)):
        output_dirs = [Path(output_dirs)]

    for i, model_path in enumerate(model_paths):
        model = ModelType.load(model_path)
        model_basename = os.path.basename(os.path.normpath(model_path))

        test_inputs = None
        test_outputs = None
        provider = "unknown"

        # 1. Use raw_input as test_inputs if specified;
        if cmp_cfg["raw_input"] is not None:
            EXEC_LOG.info(
                "Using raw input from {} as oracle".format(cmp_cfg["raw_input"])
            )
            test_inputs = pickle.load(Path(cmp_cfg["raw_input"]).open("rb"))
            assert isinstance(
                test_inputs, dict
            ), "Raw input type should be Dict[str, np.ndarray]"
            provider = "raw input from {}".format(cmp_cfg["raw_input"])

        # 2. Otherwise, use existing or generated oracle;
        if test_inputs is None:
            oracle_path = None
            # 1. check if we can directly use oracle from `oracle.pkl`
            if "auto" == cmp_cfg["oracle"]:
                oracle_path = model_path.replace(model_basename, "oracle.pkl")
                if not os.path.exists(oracle_path):
                    oracle_path = None
            elif cmp_cfg["oracle"] is not None:
                oracle_path = cmp_cfg["oracle"]

            if oracle_path is not None:
                EXEC_LOG.info("Using oracle from {}".format(oracle_path))
                res = pickle.load(Path(oracle_path).open("rb"))
                test_inputs = res["input"]
                test_outputs = res["output"]
                provider = res["provider"]

        if test_inputs is None:
            EXEC_LOG.info("Generating input data from BackendFactory.make_random_input")
            test_inputs = BackendFactory.make_random_input(model.input_like)
            provider = f"random inputs"

        # the oracle might not have oracle outputs.
        oracle = Oracle(test_inputs, test_outputs, provider)
        testcase = TestCase(model, oracle)

        this_fac = BackendFactory.init(
            cfg["backend"]["type"],
            target=cfg["backend"]["target"],
            optmax=cfg["backend"]["optmax"],
        )

        output_dir = None if output_dirs is None else output_dirs[i]
        verify_testcase(
            cmp_cfg,
            this_fac,
            testcase,
            output_dir,
        )


if __name__ == "__main__":
    main()



import logging
import random
import time
import traceback
from abc import abstractmethod
from typing import List, Optional, Set, Tuple, Type

import z3

from nnsmith.abstract.arith import *
from nnsmith.abstract.dtype import *
from nnsmith.abstract.op import (
    AbsOpBase,
    AbsTensor,
    Expand,
    Placeholder,
    concretize_op,
    rank_all,
)
from nnsmith.error import ConstraintError, SanityCheck
from nnsmith.gir import GraphIR, InstExpr, InstIR
from nnsmith.logging import MGEN_LOG, SMT_LOG
from nnsmith.util import HAS_PYGRAPHVIZ, set_seed, viz_dot


class RequiredDimNotFound(Exception):
    pass


def concretize_graph(ir: GraphIR, model: z3.ModelRef) -> GraphIR:
    return ir.concretize(model)


class BaseGen:
    def __init__(
        self,
        opset,
        seed=None,
        forward_prob=None,
        concr_ph_dim_rng=(1, 64),
        max_elem_per_tensor=2**16,
        rank_choices=None,
        dtype_choices=None,
    ):
        assert len(opset) > 0, "opset must not be empty"
        if seed is not None:
            set_seed(seed)

        self.seed = seed
        self.op_candidates = opset
        self.ir = GraphIR()
        self.monotonic_placeholder_id = 0

        # Names of current placeholders
        self.placeholders: List[str] = []
        # for all (including newly created tmp) placeholders
        self.forward_prob = 0.5 if forward_prob is None else forward_prob
        self.concr_ph_dim_rng = concr_ph_dim_rng
        self.max_elem_per_tensor = max_elem_per_tensor
        self.rank_choices = rank_choices if rank_choices else rank_all()
        # analyze the dtypes used by the opset
        dtype_top = set()
        for op in opset:
            dtype_top.update({dt for dtc in op.in_dtypes + op.out_dtypes for dt in dtc})

        self.dtype_choices = (
            [
                dt if isinstance(dt, DType) else DType.from_str(dt)
                for dt in dtype_choices
            ]
            if dtype_choices
            else DTYPE_GEN_ALL
        )

        self.dtype_choices = list(dtype_top.intersection(self.dtype_choices))
        assert len(self.dtype_choices) > 0, "dtype_choices must not be empty"
        assert len(self.rank_choices) > 0, "rank_choices must not be empty"

    def random_rank(self):
        return random.choice(self.rank_choices)

    def tensor_type_constraints(
        self, atensor: AbsTensor
    ) -> List[Union[z3.BoolRef, bool]]:
        return [atensor.nelement() <= self.max_elem_per_tensor]

    @abstractmethod
    def assume(self, c: Union[z3.BoolRef, bool]):
        pass

    def make_symbolic_placeholder(self, rank, dtype=None) -> Placeholder:
        syms = self.new_syms(
            [f"ph{self.monotonic_placeholder_id}_{k}" for k in range(rank)]
        )
        ph = Placeholder(
            AbsTensor(
                shape=syms,
                dtype=dtype if dtype is not None else self.random_dtype_gen(),
            )
        )
        self.monotonic_placeholder_id += 1
        return ph

    def make_random_concrete_placeholder(self, rank, dtype=None):
        l, r = self.concr_ph_dim_rng
        shape = []
        product = 1
        for _ in range(rank):
            v = random.randint(l, r)
            if product * v > self.max_elem_per_tensor:
                v = 1
            shape.append(v)
            product *= v

        random.shuffle(shape)  # shuffle

        ph = Placeholder(
            AbsTensor(
                shape=shape,
                dtype=dtype if dtype is not None else self.random_dtype_gen(),
            )
        )
        return ph

    def random_dtype_gen(self):
        # more floats than ints.
        # ~ in DTYPE_GEN_ALL and in self.dtype_choices
        dtypes = [dt for dt in DTYPE_GEN_ALL if dt in self.dtype_choices]
        assert (
            len(dtypes) > 0
        ), "Empty INTERSECT(DTYPE_GEN_ALL, dtype_choices). Please relax dtype_choices."

        wts = [1] * len(dtypes)
        for dt in DTYPE_GEN_FLOATS:
            if dt in dtypes:
                wts[dtypes.index(dt)] = 4
        for dt in DTYPE_GEN_INTS:
            if dt in dtypes:
                wts[dtypes.index(dt)] = 1
        return random.choices(dtypes, weights=wts)[0]

    def new_sym(self, name):
        return z3.Int(name)

    def new_syms(self, names):
        return [self.new_sym(name) for name in names]

    def insert_init_ph_node(self, ph: Placeholder) -> InstIR:
        inst = self.forward_insert_node(ph, [])

        for c in ph.ttype.sym_gt_conc_ge_zero():
            self.assume(c)

        return inst

    @abstractmethod
    def try_forward_insert_at(self, node: AbsOpBase, input_vars: List[str]) -> bool:
        raise NotImplementedError

    @abstractmethod
    def try_occupy_placeholder(self, node: AbsOpBase, phvars: List[str]) -> bool:
        raise NotImplementedError

    @abstractmethod
    def make_concrete(self) -> GraphIR:
        raise NotImplementedError

    def extra_exit_check(self, max_node_size) -> bool:
        """
        Returns:
            bool: add more checks to determine whether to exit the generation.
        """
        return False

    def num_op(self) -> int:
        # exclude placeholders.
        return self.ir.n_compute_inst()

    def try_insert(self):
        node_t = self.pick_next_op_type()
        self.try_insert_node_type(node_t)

    def abstract_gen(self, max_node_size=10, max_gen_millisec=2000):
        z3.set_param("timeout", max_gen_millisec // 3)

        assert max_node_size > 0, "max_node_size must be positive"

        init_time = time.time()

        # starts generation.
        while (
            time.time() - init_time < max_gen_millisec / 1000
            and self.num_op() < max_node_size
        ):
            if self.extra_exit_check(max_node_size):
                break
            self.try_insert()

        # init graph placeholders
        SanityCheck.gt(len(self.placeholders), 0)

        def determine_ph_type(ph: str, to_input: bool):
            SanityCheck.true(ph in self.placeholders)
            ph_inst_id, _ = InstIR.var_inst_idx(ph)
            ph_inst = self.ir.find_inst_by_id(ph_inst_id)
            if to_input:
                ph_inst.iexpr.op = ph_inst.iexpr.op.to_input()
            else:
                ph_inst.iexpr.op = ph_inst.iexpr.op.to_const()

        determine_ph_type(self.placeholders[0], True)  # At lease make one input.
        for ph in self.placeholders[1:]:
            determine_ph_type(ph, random.randint(0, 1))

    def pick_next_op_type(self):
        return random.choice(self.op_candidates)

    def forward_insert_node(self, node: AbsOpBase, input_vars: List[str]) -> InstIR:
        new_inst = self.ir.add_inst(InstExpr(op=node, args=input_vars))

        if isinstance(node, Placeholder):
            # Add placeholder's return varname.
            self.placeholders.append(new_inst.retval())

        return new_inst

    def backward_insert_node(
        self, node, input_vars: List[str], ph_to_replace: List[str]
    ) -> InstIR:
        new_inst = self.forward_insert_node(node, input_vars=input_vars)

        # replace all uses of ph_to_replace
        # and delete the unused placeholders.
        for ph, rv in zip(ph_to_replace, new_inst.retvals()):
            self.ir.replace_alluse(ph, rv)
            ph_inst_id, _ = InstIR.var_inst_idx(ph)
            ph_inst = self.ir.find_inst_by_id(ph_inst_id)
            self.ir.remove_unused(ph_inst)
            self.placeholders.remove(ph)

        return new_inst

    def try_forward_insert(self, op: AbsOpBase) -> bool:
        n_inp = len(op.inp_ranks)
        dim_spec_list = []

        if op.same_inp_dims:  # find `n_inp` under the same input shapes.
            rank_set = set(op.inp_ranks[0])

            for ranks in op.inp_ranks[1:]:
                rank_set.intersection_update(set(ranks))

            SanityCheck.ge(len(rank_set), 1)

            final_dim = random.choice(list(rank_set))
            dim_spec_list = [(final_dim,)] * n_inp
        else:  # inputs have different dimension sizes.
            dim_spec_list = op.inp_ranks

        varnames = self.pick_var_group(
            dim_spec_list,
            op.in_dtypes,
        )

        if self.try_forward_insert_at(op, varnames):
            return True

        return False

    def try_backward_insert(self, op: AbsOpBase):
        # we know that: Y = op(X)
        # S1 - select Y: Y must be a placeholder; (this also means the graph must start w/ a placeholder)
        phvars = self.pick_var_group(
            op.out_ranks,
            op.out_dtypes,
            var_candidates=[
                name
                for name in self.placeholders
                if not isinstance(op, Expand)
                or self.ir.vars[name].ndims < op.expand_last_dim
            ],
        )

        if self.try_occupy_placeholder(op, phvars):
            return True

        return False

    def try_insert_node_type(
        self, node_t: Type[AbsOpBase], max_tensor_pick_time=3
    ) -> bool:
        MGEN_LOG.debug(
            f"@[Node #{self.ir.n_inst()}] <-- trying to insert node type {node_t.__name__}"
        )

        try:
            for _ in range(max_tensor_pick_time):
                # should recreate a new instance since some attributes (like axis) should be initialized for each pick
                op_param_n = node_t.get_num_var_param()
                op_id = self.ir.n_inst()
                op_params = [
                    self.new_sym("op%s_%s" % (op_id, k)) for k in range(op_param_n)
                ]

                op: AbsOpBase = node_t(*op_params)

                if random.uniform(0, 1) < self.forward_prob:
                    if self.try_forward_insert(op):
                        return True
                else:
                    if self.try_backward_insert(op):
                        return True
        except RequiredDimNotFound:
            if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
                MGEN_LOG.debug(traceback.format_exc())
            return False
        except ConstraintError:
            if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
                MGEN_LOG.debug(traceback.format_exc())
            return False

        return False

    def filter_rank_dtype(self, ndims, dtype, candidates: List[str]):
        cans = candidates

        cans = list(
            filter(  # filter with ndim
                lambda vname: self.ir.vars[vname].ndims in ndims, cans
            )
        )
        if len(cans) == 0:
            raise RequiredDimNotFound(f"Cannot find candidate to sat rank of {ndims}.")

        if dtype is not None:
            cans = list(
                filter(  # filter with dtype
                    lambda vname: self.ir.vars[vname].dtype == dtype, cans
                )
            )
            if len(cans) == 0:
                raise RequiredDimNotFound(
                    f"Cannot find candidate to sat rank of {ndims} and dtype {dtype}."
                )

        return cans

    def pick_var_group(
        self,
        ndim_list: List[Set[int]],
        dtype_combs_spec: List[Tuple[DType, ...]],
        var_candidates: Optional[List[str]] = None,
    ) -> List[str]:
        """Randomly pick a group of variables that satisfy one of the `dtype_combs_spec` and `ndim_list`.
        Returns:
            List[str]: Satisfiable group of variable names.
        """

        if var_candidates is None:
            var_candidates = list(self.ir.vars.keys())

        # check if can gen var group data types:
        dtypes_in_ir = set([self.ir.vars[vname].dtype for vname in var_candidates])
        if dtypes_in_ir.isdisjoint(set(DTYPE_GEN_ALL)):
            raise RequiredDimNotFound(
                f"DType unsat in IR (possibly due to complex64/128 dtypes)."
            )

        abs_tensor_candidates = []
        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            for cand in var_candidates:
                MGEN_LOG.debug(
                    f"Candidate: {cand}: {self.ir.vars[cand].dtype} ~ {self.ir.vars[cand].ndims}"
                )
            MGEN_LOG.debug(f"Input data ranks candidates: {ndim_list}")
            MGEN_LOG.debug(f"Input data types candidates: {dtype_combs_spec}")

        viable_dtypes = []
        for i, ndims in enumerate(ndim_list):
            viable_dtypes.extend(
                [
                    self.ir.vars[vname].dtype
                    for vname in self.filter_rank_dtype(
                        ndims=ndims, dtype=None, candidates=var_candidates
                    )
                ]
            )
        # only use dtypes currently available after ndim filtering
        dtype_combs = [
            comb for comb in dtype_combs_spec if all(dt in viable_dtypes for dt in comb)
        ]
        if len(dtype_combs) == 0:
            raise RequiredDimNotFound(
                f"No viable candidates: rank within {ndim_list} and dtype within {dtype_combs_spec}."
            )
        dtype_comb = random.choice(dtype_combs)
        for i, ndims in enumerate(ndim_list):
            candidates = self.filter_rank_dtype(
                ndims=ndims, dtype=dtype_comb[i], candidates=var_candidates
            )
            abs_tensor_candidates.append(random.choice(candidates))

        return abs_tensor_candidates


def check_sat(solver: z3.Solver, *assumptions) -> z3.CheckSatResult:
    start = time.time()

    if SMT_LOG.getEffectiveLevel() <= logging.DEBUG:
        if solver.assertions():
            SMT_LOG.debug(
                f"existing constraints: {', '.join(map(str, solver.assertions()))}"
            )
        if assumptions:
            SMT_LOG.debug(f"new constraints: {', '.join(map(str, assumptions))}")

    cres = solver.check(*assumptions)

    if SMT_LOG.getEffectiveLevel() <= logging.DEBUG:
        SMT_LOG.debug(
            f"{cres} <-- checking time: {int((time.time() - start) * 1000)} ms"
        )

        if cres == z3.unsat:
            SMT_LOG.debug(f"Unsat core: {solver.unsat_core()}")

    return cres


def set_z3_state(seed=None):
    z3.set_param(
        "smt.phase_selection",
        5,
        "smt.arith.random_initial_value",
        True,
        "smt.random_seed",
        seed,
        "sat.random_seed",
        seed,
        "sat.phase",
        "random",
        "memory_max_size",
        50 * 1024,  # MB
    )


class SymbolicGen(BaseGen):
    def __init__(
        self,
        opset,
        seed=None,
        init_fp=False,
        symbolic_init=True,
        **kwargs,
    ):
        super().__init__(opset, seed, **kwargs)
        if seed is not None:
            set_z3_state(seed)

        self.solver = z3.Solver()
        self.last_solution: Optional[z3.ModelRef] = None

        # Insert the first node.
        if symbolic_init:
            ph = self.make_symbolic_placeholder(
                self.random_rank(), dtype=DType.float32 if init_fp else None
            )
        else:
            ph = self.make_random_concrete_placeholder(
                self.random_rank(), dtype=DType.float32 if init_fp else None
            )

        self.insert_init_ph_node(ph)
        for pred in self.tensor_type_constraints(ph.ttype):
            self.assume(pred)

    def assume(self, c: z3.BoolRef):
        self.solver.add(c)

    def check_sat(self, *assumptions):
        cres = check_sat(self.solver, *assumptions)
        if cres == z3.sat:
            self.last_solution = self.solver.model()
        return cres

    def try_forward_insert_at(self, node: AbsOpBase, input_vars: List[str]) -> bool:
        itensors = [self.ir.vars[vname] for vname in input_vars]
        constraints = node.checked_requires(itensors)

        if SMT_LOG.getEffectiveLevel() <= logging.DEBUG:
            SMT_LOG.debug(f"---> Trying to solve: {node} ~ {constraints}")

        # make a copy
        otensors = node.checked_type_transfer(itensors)

        for aten in otensors:
            for c in aten.gt_zero():
                constraints.append(c)

        # limit output tensor size
        for aten in otensors:
            constraints.extend(self.tensor_type_constraints(aten))

        check_res = self.check_sat(*constraints)

        if check_res != z3.sat:
            return False

        for c in constraints:
            self.assume(c)

        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            MGEN_LOG.debug(f">> Forward insert: {node}")
            MGEN_LOG.debug(f"\tinputs:  {itensors}")
            MGEN_LOG.debug(f"\toutputs: {otensors}")

        node.bind_input_like(itensors)
        node.bind_output_like(otensors)

        self.forward_insert_node(node, input_vars)
        return True

    def try_occupy_placeholder(self, node: AbsOpBase, phvars: List[str]) -> bool:
        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            MGEN_LOG.debug(
                f"---> Trying to occupy placeholder: {phvars} for node {node}"
            )
        # S2 - create X: X can be
        #                   - a new placeholder (fallback)
        #                   - an existing alive shape

        otensors = [self.ir.vars[name] for name in phvars]

        # S2.2: try to reuse some existing outputs;
        # TODO: allow reuse existing alive shapes
        # n_inps = len(node.inp_ranks)
        # max_try = 2
        # n_reuse = n_inps - 1
        # while n_reuse > 0 and max_try > 0:
        #     # TODO...
        #     max_try -= 1
        #     n_reuse -= 1

        # S2.2: reusing outputs failed. as a fallback, promote all free vars to placeholders.
        phs_as_op_inputs: List[Placeholder] = []
        constraints = []
        for rank, dtype in node.deduct_inp_ranks_and_dtype(otensors):
            # oversample rank 4 tensors as they may be more important
            ph = self.make_symbolic_placeholder(
                rank if rank != -1 else self.random_rank(), dtype=dtype
            )
            phs_as_op_inputs.append(ph)
            constraints.extend(ph.ttype.gt_zero())
            constraints.extend(self.tensor_type_constraints(ph.ttype))

        itensors = [p.ttype for p in phs_as_op_inputs]
        constraints.extend(node.checked_requires(itensors))
        inferred_otensors = node.checked_type_transfer(itensors)

        for i, shape in enumerate(inferred_otensors):
            constraints.extend(shape.eq(otensors[i]))
            constraints.extend(shape.gt_zero())

        check_res = self.check_sat(*constraints)

        if check_res != z3.sat:
            return False

        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            MGEN_LOG.debug(f">> Backward insert: {node}")
            MGEN_LOG.debug(f"\tinputs:  {phs_as_op_inputs}")

        for c in constraints:
            self.assume(c)

        # succ.
        input_vars = []

        for ph in phs_as_op_inputs:
            inst = self.forward_insert_node(ph, [])
            input_vars.append(inst.retval())

        node.bind_input_like(itensors)
        node.bind_output_like(inferred_otensors)

        self.backward_insert_node(node, input_vars, phvars)

        return True

    def make_concrete(self) -> GraphIR:
        SanityCheck.gt(len(self.ir.insts), 0, "Empty graph!")
        SanityCheck.not_none(self.last_solution, "Run check_sat first!")
        self.ir.concretize(self.last_solution)
        return self.ir


class ConcolicGen(BaseGen):
    """Different from SymbolicGen, the graph after an insertion is `concrete` in ConcolicGen.
    However, each step when inserting a node, we symbolically find a satisfiable solution for it."""

    def __init__(
        self,
        opset,
        seed=None,
        init_fp=False,
        **kwargs,
    ):
        super().__init__(opset, seed, **kwargs)
        if seed is not None:
            set_z3_state(seed)

        # Insert the first node.
        self.insert_init_ph_node(
            self.make_random_concrete_placeholder(
                self.random_rank(), dtype=DType.float32 if init_fp else None
            )
        )

    def try_forward_insert_at(self, node: AbsOpBase, input_vars: List[str]) -> bool:
        solver = z3.Solver()

        itensors = [self.ir.vars[vname] for vname in input_vars]
        constraints = node.checked_requires(itensors)

        if SMT_LOG.getEffectiveLevel() <= logging.DEBUG:
            SMT_LOG.debug(f"---> Trying to solve: {node} ~ {constraints}")

        # make a copy
        otensors = node.checked_type_transfer(itensors)

        for aten in otensors:
            for c in aten.sym_gt_conc_ge_zero():
                constraints.append(c)

        check_res = check_sat(solver, *constraints)

        if check_res != z3.sat:
            return False

        # materialize otensors and attributes.
        node = concretize_op(node, solver.model())
        otensors = node.checked_type_transfer(itensors)

        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            MGEN_LOG.debug(f">> Forward insert: {node}")
            MGEN_LOG.debug(f"\tinputs:  {itensors}")
            MGEN_LOG.debug(f"\toutputs: {otensors}")

        # Shape checker.
        # NOTE: No need to check input shape as they are already checked for being in the graph.
        for i, ten in enumerate(otensors):
            if not all(self.tensor_type_constraints(ten)):
                MGEN_LOG.debug(f"{i}-th output type constraint failed: {ten}")
                return False

        node.bind_input_like(itensors)
        node.bind_output_like(otensors)

        self.forward_insert_node(node, input_vars)
        return True

    def try_occupy_placeholder(self, node: AbsOpBase, phvars: List[str]) -> bool:
        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            MGEN_LOG.debug(
                f"---> Trying to occupy placeholder: {phvars} for node {node}"
            )

        # TODO: In backward insertion, reusing existing tensors is not implemented.

        # Concrete tensors.
        solver = z3.Solver()

        otensors = [self.ir.vars[name] for name in phvars]
        phs_as_op_inputs: List[Placeholder] = []
        constraints = []
        for rank, dtype in node.deduct_inp_ranks_and_dtype(otensors):
            # oversample rank 4 tensors as they may be more important
            ph = self.make_symbolic_placeholder(
                rank if rank != -1 else self.random_rank(), dtype=dtype
            )
            phs_as_op_inputs.append(ph)
            constraints.extend(ph.ttype.sym_gt_conc_ge_zero())

        itensors = [p.ttype for p in phs_as_op_inputs]
        constraints.extend(node.checked_requires(itensors))
        inferred_otensors = node.checked_type_transfer(itensors)

        for i, shape in enumerate(inferred_otensors):
            constraints.extend(shape.eq(otensors[i]))
            constraints.extend(shape.sym_gt_conc_ge_zero())

        check_res = check_sat(solver, *constraints)

        if check_res != z3.sat:
            return False

        if MGEN_LOG.getEffectiveLevel() <= logging.DEBUG:
            MGEN_LOG.debug(f">> Backward insert: {node}")
            MGEN_LOG.debug(f"\tinputs:  {phs_as_op_inputs}")

        model = solver.model()
        # succ.
        itensors = []
        for i, ph in enumerate(phs_as_op_inputs):
            # materialize ph
            phs_as_op_inputs[i] = concretize_op(ph, model)
            itensors.append(phs_as_op_inputs[i].ttype)

        # Input Shape checker.
        # NOTE: No need to check output because they are already in the graph thus valid.
        for i, ten in enumerate(itensors):
            if not all(self.tensor_type_constraints(ten)):
                MGEN_LOG.debug(f"{i}-th input type constraint failed: {ten}")
                return False

        node = concretize_op(node, model)
        node.bind_input_like(itensors)
        node.bind_output_like(otensors)

        # Apply insertion.
        input_vars = []
        for ph in phs_as_op_inputs:
            inst = self.forward_insert_node(ph, [])
            input_vars.append(inst.retval())

        self.backward_insert_node(node, input_vars, phvars)

        return True

    def assume(self, c: bool):
        # semantically equivalent to `assert c`.
        ConstraintCheck.true(c, "Assumption failed")

    def make_concrete(self) -> GraphIR:
        return self.ir


def model_gen(
    opset: Set[Type[AbsOpBase]],
    method: str = "symbolic",
    max_nodes=5,
    seed=None,
    timeout_ms=10000,
    **kwargs,
):
    assert max_nodes > 0, "max_nodes must >= 1"

    if "symbolic" == method or "symbolic-sinit" == method:
        gen = SymbolicGen(opset, seed, symbolic_init=True, **kwargs)
    elif "symbolic-cinit" == method:
        gen = SymbolicGen(opset, seed, symbolic_init=False, **kwargs)
    elif "concolic" == method:
        gen = ConcolicGen(opset, seed, **kwargs)
    else:
        raise ValueError(f"Unknown method {method}. Try `symbolic` or `concolic`.")

    gen.abstract_gen(max_node_size=max_nodes, max_gen_millisec=timeout_ms)

    return gen


def viz(ir: GraphIR, filename: str = None):
    if HAS_PYGRAPHVIZ:
        viz_dot(ir.to_dot(), filename)



import logging

import tvm
from multipledispatch import dispatch
from tvm import relay

from nnsmith.backends import BackendFactory
from nnsmith.materialize.onnx import ONNXModel

logging.getLogger("te_compiler").disabled = True
logging.getLogger("autotvm").disabled = True


def list_eq(a, b):
    if len(a) != len(b):
        return False
    for i in range(len(a)):
        if a[i] != b[i]:
            return False
    return True


class TVMFactory(BackendFactory):
    def __init__(self, target="cpu", optmax=True, executor="graph", **kwargs) -> None:
        super().__init__(target, optmax, **kwargs)
        # WARNING: setting opt_level 4 sometimes causes false alarms
        # as in this level fast_math is enabled where slight numerical
        # inconsistency is allowed and outputs for UB-input may differ.
        self.opt_level = 4 if optmax else 0
        if target == "cpu":
            self.tvm_target = tvm.target.Target("llvm")
        else:
            tvm_possible_targets = tvm.target.Target.list_kinds()
            assert (
                target in tvm_possible_targets
            ), f"Unknown target {target}. Possible targets are {tvm_possible_targets}"
            self.tvm_target = tvm.target.Target(target)

        self.executor_mode = executor

    def get_device(self):
        dev_cand = self.tvm_target.export()["keys"]
        assert len(dev_cand) > 0, f"No viable device found for {self.tvm_target}"
        if "cuda" in dev_cand:
            return tvm.cuda()
        if "rocm" in dev_cand:
            return tvm.rocm()
        if "cpu" in dev_cand:
            return tvm.cpu()
        return tvm.device(dev_cand[0])

    @property
    def system_name(self) -> str:
        return "tvm"

    @staticmethod
    def cvt_result(output):
        """Pack output tensor(s) into a list"""
        # TODO(jinkun): may not work for nested list / dynamic shape
        assert output is not None, "Output should not be None"
        if isinstance(output, (tvm.runtime.container.ADT, list)):
            output = [r.numpy() for r in output]
        elif output is not None:
            output = [output.numpy()]
        return output

    @dispatch(ONNXModel)
    def make_backend(self, model: ONNXModel):
        onnx_model = model.native_model
        shape_dict = {name: aten.shape for name, aten in model.input_like.items()}
        mod, params = relay.frontend.from_onnx(
            onnx_model, shape_dict, freeze_params=True
        )
        mod = relay.transform.InferType()(mod)

        with tvm.transform.PassContext(opt_level=self.opt_level):
            executor = relay.build_module.create_executor(
                self.executor_mode, mod, self.get_device(), self.tvm_target, params
            ).evaluate()

        def closure(inputs):
            output = executor(**inputs)
            output = self.cvt_result(output)
            return dict(zip(model.output_like.keys(), output))

        return closure

    @property
    def version(self) -> str:
        return tvm.__version__



import multiprocessing as mp
import os
import pickle
import subprocess
import tempfile
from copy import deepcopy
from pathlib import Path


def analyze_lcov(lcov_data):
    file_covs = lcov_data.split("end_of_record\n")[:-1]
    ret = {
        # 'filename': {
        #    'lines': [identifiers],
        #    'branches': [identifiers],
        #    'lf': # lines total,
        #    'bf': # branches total,
        # }
    }

    # SF: source file
    # sperated by end_of_record

    # FN: <line number>,<function name>

    for file_cov in file_covs:
        cov_lines = file_cov.split("\n")
        assert cov_lines[0].startswith("SF:")
        filename = cov_lines[0][3:]

        lines = set()
        branches = set()

        # LH: # line hits
        # FNH: # function hits | note we consider source-level functions. e.g., templates are one function.
        # BRH: # branch hits
        # NOTE: LH might (slightly) != len(lines). we only consider len(lines).

        n_line_total = 0  # LF: # lines
        n_branch_total = 0  # BRF: # branches

        lf = 0
        brf = 0

        for cov_line in cov_lines[1:]:
            cov_line = cov_line.rstrip("\n")
            if cov_line.startswith("DA:"):
                # DA: <line number>,<execution count> for each instrumented line
                line_number, exec_count = cov_line[3:].split(",")
                if exec_count != "0":
                    lines.add(int(line_number))
                n_line_total += 1
            elif cov_line.startswith("BRDA:"):
                # BRDA: <line number>,<block number>,<branch number>,<taken>
                line_number, block_number, branch_number, taken = cov_line[5:].split(
                    ","
                )
                if taken != "-" and taken != "0":
                    branches.add(line_number + ":" + block_number + ":" + branch_number)
                n_branch_total += 1
            # elif cov_line.startswith('LH:'):
            #     n_line_hit = int(cov_line[3:])
            elif cov_line.startswith("LF:"):
                lf = int(cov_line[3:])
            # elif cov_line.startswith('FNH:'):
            #     n_func_hit = int(cov_line[4:])
            elif cov_line.startswith("BRF:"):
                brf = int(cov_line[4:])
            else:
                pass
        # # BRF might be even smaller than branches you hit. we conservatively use n_branch_total.
        n_branch_total = max(brf, n_branch_total)
        # Similarly
        n_line_total = max(lf, n_line_total)

        assert (
            len(lines) <= n_line_total
        ), f"{len(lines)} <= {n_line_total} in {filename}"
        assert (
            len(branches) <= n_branch_total
        ), f"{len(branches)} <= {n_branch_total} in {filename}"

        ret[filename] = {
            "lines": lines,
            "branches": branches,
            "lf": n_line_total,
            "bf": n_branch_total,
        }

    return ret


def get_cmd_output(cmd_list):
    return subprocess.run(cmd_list, stdout=subprocess.PIPE, check=True).stdout.decode()


def check_profraw(fname: str):
    for n in fname.split(".")[:-1]:
        if not n.isdigit():
            return False
    return True


def merge_cov(current, rhs, hint=""):
    # {
    #     'lines': lines,
    #     'branches': branches,
    #     'lf': n_line_total,
    #     'bf': n_branch_total,
    # }

    for k in set(current.keys()).union(set(rhs.keys())):
        if k not in current:
            current[k] = rhs[k]
            continue
        elif k not in rhs:
            continue
        else:
            current[k]["lines"] = current[k]["lines"].union(rhs[k]["lines"])
            current[k]["branches"] = current[k]["branches"].union(rhs[k]["branches"])

            if current[k]["lf"] != rhs[k]["lf"]:
                print(
                    f'[WARNING] total line {current[k]["lf"]} != {rhs[k]["lf"]} in {k} '
                    + hint
                )
                current[k]["lf"] = max(current[k]["lf"], rhs[k]["lf"])

            if current[k]["bf"] != rhs[k]["bf"]:
                print(
                    f'[WARNING] total branch {current[k]["bf"]} != {rhs[k]["bf"]} in {k} '
                    + hint
                )
                current[k]["bf"] = max(current[k]["bf"], rhs[k]["bf"])

    return current


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--root", type=str, required=True, help="Folder to all the tests."
    )
    parser.add_argument(
        "--llvm-config-path",
        type=str,
        required=True,
        help="Path to `llvm-config` whose version must be consistent with the compiled coverage.",
    )
    parser.add_argument(
        "--instrumented-libs",
        type=str,
        required=True,
        nargs="+",
        help="List of instrumentated libraris (.so).",
    )
    parser.add_argument(
        "--parallel",
        type=int,
        default=8,
        help="Number of parallel jobs for processing profraws.",
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        required=True,
        help="How many models evaluted in each profraw?",
    )

    args = parser.parse_args()

    assert os.path.exists(args.root)
    cov_root = os.path.join(args.root, "coverage")
    assert os.path.exists(args.llvm_config_path)
    llvm_bindir = get_cmd_output([args.llvm_config_path, "--bindir"]).strip()
    assert os.path.exists(llvm_bindir)

    llvm_profdata = os.path.join(llvm_bindir, "llvm-profdata")
    assert os.path.exists(llvm_profdata), llvm_profdata
    llvm_cov = os.path.join(llvm_bindir, "llvm-cov")
    assert os.path.exists(llvm_cov), llvm_cov

    if args.instrumented_libs:
        lib_expr = ""
        for lib in args.instrumented_libs:
            assert os.path.exists(lib), f"{lib} does not exist!"
            lib_expr += f" -object {os.path.realpath(lib)} "

    profraws = [p for p in Path(cov_root).rglob("*.profraw") if check_profraw(p.name)]
    # rank by
    profraws.sort(key=lambda f: float(f.name.rsplit(".", 1)[0]))

    merged_cov_path = os.path.join(cov_root, "merged_cov.pkl")

    with tempfile.TemporaryDirectory() as tmpdirname:

        def process_one_profraw(profraw_path: Path):
            profdata_path = os.path.join(tmpdirname, profraw_path.name + ".profdata")
            # llvm-prof-data
            profdata_cmd = (
                f"{llvm_profdata} merge -sparse {profraw_path} -o {profdata_path}"
            )
            assert 0 == os.system(profdata_cmd), profdata_cmd
            llvmcov_cmd = f"{llvm_cov} export -instr-profile={profdata_path} -format=lcov {lib_expr}"
            print(llvmcov_cmd)
            lcov_data = get_cmd_output(llvmcov_cmd.split(" "))
            return analyze_lcov(lcov_data)

        merged_cov = {
            # 'time': time point. { # ! Not time duration.
            #   'n_model': # how many models got successfully executed?
            #   'merged_cov': coverage dictionary.
            # }
        }

        current_cov = {}
        with mp.Pool(args.parallel) as pool:
            for i, res in enumerate(pool.imap(process_one_profraw, profraws)):
                profraw_name = profraws[i].name
                current_time = float(profraw_name.rsplit(".", 1)[0])

                merged_cov[current_time] = {}
                merged_cov[current_time][
                    "n_model"
                ] = args.batch_size  # inaccurate for the last.
                current_cov = merge_cov(current_cov, res, f"merging {profraw_name}")
                merged_cov[current_time]["merged_cov"] = deepcopy(current_cov)

        with open(merged_cov_path, "wb") as fp:
            pickle.dump(merged_cov, fp)



from typing import Dict

import numpy as np
from numpy import testing


def assert_allclose(
    actual: Dict[str, np.ndarray],
    desired: Dict[str, np.ndarray],
    actual_name: str,
    oracle_name: str,
    equal_nan=False,
    rtol=1e-2,
    atol=1e-3,
):
    akeys = set(actual.keys())
    dkeys = set(desired.keys())
    if akeys != dkeys:
        raise KeyError(f"{actual_name}: {akeys} != {oracle_name}: {dkeys}")

    for key in akeys:
        lhs = actual[key]
        rhs = desired[key]

        # check if lhs is np.ndarray
        if not isinstance(lhs, np.ndarray):
            raise TypeError(f"{actual_name}[{key}] is not np.ndarray but {type(lhs)}")

        # check if rhs is np.ndarray
        if not isinstance(rhs, np.ndarray):
            raise TypeError(f"{oracle_name}[{key}] is not np.ndarray but {type(rhs)}")

        testing.assert_allclose(
            lhs,
            rhs,
            equal_nan=equal_nan,
            rtol=rtol,
            atol=atol,
            err_msg=f"{actual_name} != {oracle_name} at {key}",
        )



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Tuple, Union

from z3 import ModelRef

from nnsmith.abstract.op import (
    AbsOpBase,
    AbsTensor,
    Constant,
    Input,
    Placeholder,
    concretize_op,
)
from nnsmith.logging import CORE_LOG


@dataclass
class InstExpr:
    op: Union[AbsOpBase, Placeholder]
    args: Tuple[str]

    def __str__(self):
        return f"{self.op}({', '.join(self.args)})"

    def n_input(self):
        assert len(self.args) == self.op.n_input()
        return len(self.args)

    def n_output(self):
        return self.op.n_output()


def _make_new_id_from_used(used_ids: Set[int]) -> int:
    for i in range(2**16):
        if i not in used_ids:
            return i

    raise RuntimeError("Cannot find a new id within [0, 2**16)")


def id_maker(fallback, irctx: Union["GraphIR", List["InstIR"]] = None):
    if isinstance(irctx, GraphIR):
        return _make_new_id_from_used({inst.identifier for inst in irctx.insts})
    elif isinstance(irctx, list):
        return _make_new_id_from_used({inst.identifier for inst in irctx})
    else:
        return fallback


def id_checker(id):
    assert id >= 0, "Identifier must be non-negative"


VARNAME_SPLITTER = "_"


class InstIR:
    def __init__(
        self,
        iexpr: InstExpr,
        identifier: Optional[int] = None,
        irctx: Union["GraphIR", List["InstIR"]] = None,
    ) -> None:
        self.iexpr = iexpr
        self.users: List[Set[InstIR]] = [set() for _ in range(self.iexpr.n_output())]
        if identifier is None:
            self.identifier = id_maker(fallback=id(self), irctx=irctx)
        else:
            id_checker(identifier)
            self.identifier = identifier

        if irctx is not None:
            if isinstance(irctx, GraphIR):
                for arg in set(self.iexpr.args):
                    inst_id, ret_idx = InstIR.var_inst_idx(arg)
                    irctx.find_inst_by_id(inst_id).users[ret_idx].add(self)
            elif isinstance(irctx, List):
                for arg in set(self.iexpr.args):
                    inst_id, ret_idx = InstIR.var_inst_idx(arg)
                    for inst in irctx:
                        if inst.identifier == inst_id:
                            inst.users[ret_idx].add(self)

    def __str__(self):
        return (
            f"{', '.join(self.retvals())} = {self.iexpr} \t# inst id: {self.identifier}"
        )

    def no_users(self):
        return all(len(u) == 0 for u in self.users)

    def leaf_var(self) -> List[str]:
        ret = []
        for idx, users in enumerate(self.users):
            if len(users) == 0:
                ret.append(self.retval(idx))
        return ret

    def n_input(self):
        return self.iexpr.n_input()

    def n_output(self):
        return self.iexpr.n_output()

    @staticmethod
    def var_inst_idx(varname: str) -> Tuple[int, int]:
        # parse v[inst-id]_[index]
        tokens = varname[1:].split(VARNAME_SPLITTER)
        return int(tokens[0]), int(tokens[1])

    @staticmethod
    def retval_string(inst_id: int, ret_idx: int) -> str:
        return f"v{inst_id}{VARNAME_SPLITTER}{ret_idx}"

    def retval(self, index=0) -> str:
        assert index < self.n_output(), f"Only has {self.n_output()} outputs in {self}"
        return self.retval_string(self.identifier, index)

    def retvals(self) -> List[str]:
        return [self.retval(i) for i in range(self.n_output())]

    def is_user_of(self, inst: "InstIR", ret_idx: Optional[int] = None) -> bool:
        usee_names = list(inst.retvals())
        if ret_idx is not None:
            if ret_idx >= len(usee_names) or ret_idx < 0:
                raise ValueError(
                    f"Invalid ret_idx: {ret_idx}. Valid domain is [0, {len(usee_names)})"
                )
            return usee_names[ret_idx] in self.iexpr.args
        else:
            return any(u in self.iexpr.args for u in usee_names)


# -----------------------------------------------------------------
#                       Graph IR Structure
# -----------------------------------------------------------------
#       vars:    Dict[ var_name, AbsTensor      ]
#       insts:   List[cps(AbsOp, List[var_name])]
# -----------------------------------------------------------------
#                         Well-formedness
# -----------------------------------------------------------------
#  1. Return name: ${inst.id}_${index(ret)}. e.g., op[0]_0, ...
#  3. Return name is unique.
# -----------------------------------------------------------------


@dataclass
class GraphIR:
    vars: Dict[str, AbsTensor] = field(default_factory=dict)  # VarName  -> AbsTensor
    insts: List[InstIR] = field(
        default_factory=list
    )  # Index -> InstIR := <ID, InstExpr, RetNames..>

    def __str__(self) -> str:
        ret = ""
        for inst in self.insts:
            ret += f"{inst}\n"

        return ret

    def pretty(self) -> str:
        inst_remap = {inst.identifier: f"{idx}" for idx, inst in enumerate(self.insts)}

        ret = ""
        for inst in self.insts:
            pretty_args = []
            for arg in inst.iexpr.args:
                inst_id, ret_idx = InstIR.var_inst_idx(arg)
                pretty_args.append(InstIR.retval_string(inst_remap[inst_id], ret_idx))
            pretty_retvals = [
                InstIR.retval_string(inst_remap[inst.identifier], ret_idx)
                for ret_idx in range(inst.n_output())
            ]

            ret += f"{', '.join(pretty_retvals)} = {inst.iexpr.op}({', '.join(pretty_args)})"
            ret += f" \t# inst id: {inst_remap[inst.identifier]}\n"  # Comment

        return ret

    def n_inst(self) -> int:
        return len(self.insts)

    def n_compute_inst(self) -> int:
        return sum(
            1
            for inst in self.insts
            if not isinstance(inst.iexpr.op, (Input, Constant, Placeholder))
        )

    def n_var(self) -> int:
        return len(self.vars)

    def leaf_inst(self) -> List[str]:
        return [inst for inst in self.insts if inst.no_users()]

    def leaf_var(self) -> List[str]:
        # Non-leaf instructions can produce leaf variables.
        lvs = []
        for inst in self.insts:
            for lv in inst.leaf_var():
                if lv not in lvs:
                    lvs.append(lv)
        return lvs

    def input_var(self) -> List[str]:
        return [
            inst.retval() for inst in self.insts if isinstance(inst.iexpr.op, Input)
        ]

    def add_inst(self, iexpr: InstExpr) -> InstIR:
        new_inst = InstIR(iexpr, irctx=self)

        # Infer the output type if iexpr.op is not binded.
        otensors = iexpr.op.output_like
        if any([t is None for t in otensors]):
            otensors = iexpr.op.checked_type_transfer(
                [self.vars[arg] for arg in iexpr.args]
            )

        # make new values
        for ridx, abstensor in enumerate(otensors):
            vname = new_inst.retval(ridx)
            assert vname not in self.vars, "Variable name is not unique: " + vname
            self.vars[vname] = abstensor

        min_user_idx = 0
        # update users
        for arg in set(iexpr.args):
            assert arg in self.vars, "Variable not defined: " + arg
            inst_id, ret_idx = InstIR.var_inst_idx(arg)
            for idx, may_prod in enumerate(self.insts):
                if inst_id == may_prod.identifier:
                    may_prod.users[ret_idx].add(new_inst)
                    min_user_idx = max(min_user_idx, idx + 1)
                    break

        # insert it to somewhere that follows good topological order
        # i.e., right after the last arg.
        self.insts.insert(min_user_idx, new_inst)
        return new_inst

    def find_inst_by_id(self, obj_id: int) -> Optional[InstIR]:
        for inst in self.insts:
            if inst.identifier == obj_id:
                return inst
        return None

    def replace_alluse(self, oldvar: str, newvar: str, type_check=True) -> None:
        # Change one variable to another new variable.
        assert oldvar in self.vars, "oldvar not defined: " + oldvar
        assert newvar in self.vars, "newvar not defined: " + newvar
        # check type
        if (
            type_check
            and self.vars[oldvar] is not None
            and self.vars[newvar] is not None
        ):
            assert self.vars[oldvar].weak_compare(
                self.vars[newvar]
            ), f"Type mismatch: {self.vars[oldvar]} != {self.vars[newvar]}"
        # 1. replace all user site of oldvar to newvar.
        old_inst_id, old_ret_idx = InstIR.var_inst_idx(oldvar)
        old_inst = self.find_inst_by_id(old_inst_id)
        assert old_inst is not None, "oldvar not defined: " + oldvar
        for ouser in old_inst.users[old_ret_idx]:
            # change all use of oldvar to newvar
            ouser.iexpr.args = [newvar if a == oldvar else a for a in ouser.iexpr.args]
        # 2. change users of oldvar -> newvar.
        new_inst_id, new_ret_idx = InstIR.var_inst_idx(newvar)
        new_inst = self.find_inst_by_id(new_inst_id)
        new_inst.users[new_ret_idx] = old_inst.users[old_ret_idx]
        old_inst.users[old_ret_idx] = set()  # reset

    def replace_arg(self, inst: InstIR, arg_idx: int, newvar: str, type_check=True):
        # Change one variable to another new variable.
        assert newvar in self.vars, "newvar not defined: " + newvar
        assert (
            0 <= arg_idx < len(inst.iexpr.args)
        ), f"Invalid argument index {arg_idx} for {inst}"
        oldvar = inst.iexpr.args[arg_idx]

        # check type
        if (
            type_check
            and self.vars[oldvar] is not None
            and self.vars[newvar] is not None
        ):
            assert (
                self.vars[oldvar] == self.vars[newvar]
            ), f"Type mismatch: {self.vars[oldvar]} != {self.vars[newvar]}"
        # 1. replace current user site of oldvar to newvar.
        inst.iexpr.args[arg_idx] = newvar
        # 2. dec ref for oldvar
        if oldvar not in inst.iexpr.args:
            old_inst_id, old_ret_idx = InstIR.var_inst_idx(oldvar)
            old_inst = self.find_inst_by_id(old_inst_id)
            old_inst.users[old_ret_idx].remove(inst)
        # 3. inc ref for newvar
        new_inst_id, new_ret_idx = InstIR.var_inst_idx(newvar)
        new_inst = self.find_inst_by_id(new_inst_id)
        new_inst.users[new_ret_idx].add(inst)

    def remove_unused(self, inst: InstIR) -> None:
        # Remove an instruction which is deemed unused.
        assert inst in self.insts, f"Instruction not in graph: {inst}"
        assert inst.no_users(), f"{inst} has users {inst.users}."
        # remove users
        for other in self.insts:
            if other != inst:
                for users in other.users:
                    if inst in users:
                        users.remove(inst)
        # remove values
        for val in inst.retvals():
            del self.vars[val]
        # remove inst
        self.insts.remove(inst)

    def assert_wellform(self):
        # TODO: Check connectivity.
        defined = set()
        for inst in self.insts:
            for arg in inst.iexpr.args:
                assert arg in self.vars, f"Variable not defined: {arg}"
                assert arg in defined, f"Variable not defined yet: {arg}"
                # check usee.
                usee_id, ret_idx = InstIR.var_inst_idx(arg)
                usee = self.find_inst_by_id(usee_id)
                assert (
                    inst in usee.users[ret_idx]
                ), f"Use-Def chain broken: {usee} should be used by {inst}"

            for rv in inst.retvals():
                assert rv in self.vars, f"Return var not in self.vars: {rv}"
                assert rv not in defined, f"Variable defined twice: {rv}"

            # check user.
            for ret_idx, users in enumerate(inst.users):
                val = inst.retval(ret_idx)
                for user in users:
                    assert (
                        val in user.iexpr.args
                    ), f"Use-Def chain broken: {inst} should be used by {user}"

            defined.update(inst.retvals())

    def _topological_sort(self):
        defined = set()

        def swap(i, j):
            self.insts[i], self.insts[j] = self.insts[j], self.insts[i]

        ptr = 0
        while ptr < len(self.insts):
            frontier = []
            for idx in range(ptr, len(self.insts)):
                inst = self.insts[idx]
                if all(arg in defined for arg in inst.iexpr.args):
                    frontier.append(idx)
            if len(frontier) == 0:
                CORE_LOG.error(f"Bad IR:\n{self.pretty()}")
                raise RuntimeError("Cyclic dependency detected.")
            for idx in frontier:
                swap(ptr, idx)
                defined.update(self.insts[ptr].retvals())
                ptr += 1

    def _udchain_repair(self):
        for inst in self.insts:
            # Add used;
            for arg in inst.iexpr.args:
                usee_id, ret_idx = InstIR.var_inst_idx(arg)
                usee = self.find_inst_by_id(usee_id)
                usee.users[ret_idx].add(inst)
            # Remove unused;
            for ret_idx, users in enumerate(inst.users):
                val = inst.retval(ret_idx)
                for user in list(users):
                    if val not in user.iexpr.args:
                        users.remove(user)

    def wellform_repair(self):
        # 1. Repair use-def chain;
        self._udchain_repair()
        # 2. Repair topological order;
        self._topological_sort()

    def concretize(self, model: ModelRef) -> None:
        """Concretize self with a z3 model."""
        for inst in self.insts:
            # Concretize operators
            op = concretize_op(inst.iexpr.op, model)

            # Concretize output tensors;
            itensors = [self.vars[vname] for vname in inst.iexpr.args]
            otensors = op.checked_type_transfer(itensors)
            op.bind_input_like(itensors)
            op.bind_output_like(otensors)

            # Write back op to insts.
            inst.iexpr.op = op

            # Write back tensors to vars.
            for vname, tensor in zip(inst.retvals(), otensors):
                self.vars[vname] = tensor

    def to_dot(self) -> str:
        """Convert to graphviz dot format."""
        # digraph D {
        #     node [shape=Mrecord];
        #     Input [label="{Input|<f0> out}"];
        #     Add   [label="{{<f0> lhs| <f1> rhs}|Add|{<f2> out}}"];
        #     Input:f0 -> Add:f0;
        #     Input:f0 -> Add:f1;
        # }
        text = "digraph D {\n"
        text += "  node [shape=Mrecord];\n"

        def render_node(inst: InstIR):
            label = "{"
            extra = ""
            if not isinstance(inst.iexpr.op, (Input, Constant, Placeholder)):
                # Inputs
                label += "{"
                label += "|".join(
                    [f"<i{idx}> {arg}" for idx, arg in enumerate(inst.iexpr.args)]
                )
                label += "}|"
            elif isinstance(inst.iexpr.op, Input):
                extra += "fillcolor=cadetblue1,style=filled,"
            elif isinstance(inst.iexpr.op, Constant):
                extra += "fillcolor=lightpink,style=filled,"
            elif isinstance(inst.iexpr.op, Placeholder):
                extra += "fillcolor=lightgray,style=filled,"

            # Op
            label += f"{inst.iexpr.op}|".replace("{", r"\{").replace("}", r"\}")
            # Outputs
            label += "{"
            label += "|".join(
                [f"<o{idx}> {rv}" for idx, rv in enumerate(inst.retvals())]
            )

            label += r"}}"

            return f'  {inst.identifier} [label="{label}",{extra}];\n'

        # render nodes
        for inst in self.insts:
            text += render_node(inst)

        # render edges
        for inst in self.insts:
            for idx, arg in enumerate(inst.iexpr.args):
                usee_id, ret_idx = InstIR.var_inst_idx(arg)
                text += f'  {usee_id}:o{ret_idx} -> {inst.identifier}:i{idx} [label="{self.vars[arg].pretty()}"];\n'

        text += "}\n"
        return text



import pickle

import pytest

from nnsmith.abstract.dtype import DType
from nnsmith.abstract.op import MaxPool2d
from nnsmith.abstract.tensor import AbsTensor


def test_dtype_picklable():
    dtype = DType.from_str("float32")
    dtype2 = pickle.loads(pickle.dumps(dtype))
    assert dtype == dtype2


def test_abstensor_picklable():
    shape = [1, 2, 3]
    dtype = DType.from_str("float32")
    tensor = AbsTensor(shape, dtype)
    tensor2 = pickle.loads(pickle.dumps(tensor))
    assert tensor == tensor2
    assert tensor.shape == tensor2.shape
    assert tensor.dtype == tensor2.dtype


def test_absop_picklable():
    maxpool = MaxPool2d(kernel_h_size=2, kernel_w_size=2, stride=1, padding=0)
    maxpool2 = pickle.loads(pickle.dumps(maxpool))
    # TODO(@ganler): make AbsOp comparable.
    # assert maxpool == maxpool2



import os
from pathlib import Path
import argparse
from multiprocessing import Process
import time

import tensorflow as tf
from tensorflow import keras
from tensorflow.python.framework.convert_to_constants import (
    convert_variables_to_constants_v2,
)
import tf2onnx
from tqdm import tqdm
import numpy as np

from nnsmith.util import mkdir

tf.get_logger().setLevel("WARNING")  # Tensorflow made quiet.


def analyze_inputs_outputs(graph):
    ops = graph.get_operations()
    outputs_set = set(ops)
    inputs = []
    for op in ops:
        if len(op.inputs) == 0 and op.type == "Placeholder":
            inputs.append(op)
        else:
            for input_tensor in op.inputs:
                if input_tensor.op in outputs_set:
                    outputs_set.remove(input_tensor.op)
    outputs = list(outputs_set)
    # Control nodes shall not be considered.
    # input like: "import/x" -> x
    # output like: "import/Identity", "import/Identity_1" -> Identity, Identity_1
    inputs = [x.name.split("/")[-1] for x in inputs if "_control_node" not in x.name]
    outputs = [x.name.split("/")[-1] for x in outputs if "_control_node" not in x.name]
    return (inputs, outputs)


def keras2tf(model):
    full_model = tf.function(lambda x: model(x))
    freeze_shape = model.inputs[0].shape

    shape_list = []
    for v in freeze_shape:
        try:
            shape_list.append(int(v))
        except TypeError as e:
            shape_list.append(1)

    full_model = full_model.get_concrete_function(
        tf.TensorSpec(tf.TensorShape(shape_list), model.inputs[0].dtype)
    )
    frozen_func = convert_variables_to_constants_v2(full_model)

    return frozen_func.graph.as_graph_def()


def convert_tf2onnx(from_path, to_path):
    def custom_objects():
        def no_activation(x):
            return x

        def leakyrelu(x):
            return keras.activations.relu(x, alpha=0.01)

        objects = {}
        objects["no_activation"] = no_activation
        objects["leakyrelu"] = leakyrelu
        return objects

    model = keras.models.load_model(from_path, custom_objects=custom_objects())
    shape_list = []
    for v in model.inputs[0].shape:
        try:
            shape_list.append(int(v))
        except TypeError as e:
            shape_list.append(1)

    graph_def = keras2tf(model)

    with tf.Graph().as_default() as graph:
        tf.import_graph_def(graph_def)
        inps, outs = analyze_inputs_outputs(graph)

    tf2onnx.convert.from_graph_def(
        graph_def,
        input_names=[inp + f":{i}" for i, inp in enumerate(inps)],
        output_names=[out + f":{i}" for i, out in enumerate(outs)],
        opset=13,
        output_path=to_path,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--lemon_output_dir", type=str, required=True, help="Path to the folder."
    )
    parser.add_argument("--onnx_dir", type=str, required=True)
    parser.add_argument("--tlimit", type=int, default=4 * 60 * 60)
    args = parser.parse_args()

    mkdir(args.onnx_dir)

    # FORMAT: {generation time cost in seconds}, {model relative path}
    # MUST RANK by GENERATION ORDER.
    config_file = open(os.path.join(args.onnx_dir, "gentime.csv"), "w")

    time_list = []
    file_list = []
    for file in Path(args.lemon_output_dir).rglob("*/*.h5"):
        file_list.append(file)
        time_list.append(file.stat().st_mtime)

    assert len(file_list) > 0, "No files found in the folder!"
    print(f"{len(file_list)} files found in the folder.")

    time_arr = np.array(time_list)
    time_arr -= time_arr.min()
    idx = time_arr.argsort()

    time_span = time_arr.max() - time_arr.min()
    print(
        f"|T_last - T_first| = {time_span:.2f}s = {time_span / 60:.2f}min = {time_span / 3600:.2f}h"
    )
    assert (
        time_span / 3600 < 25
    ), "Are you sure your steps are correct? The time span is too long..."

    time_diffs = np.diff(np.sort(time_arr))

    cvt_start_time = time.time()

    for i in tqdm(range(len(file_list))):
        ranked_idx = idx[i]
        from_path = file_list[ranked_idx]
        to_name = os.path.split(from_path)[-1] + ".onnx"
        to_path = os.path.join(args.onnx_dir, to_name)
        try:
            tstart = time.time()
            # might crash...
            p = Process(target=convert_tf2onnx, args=(from_path, to_path))
            p.start()
            p.join()

            assert p.exitcode is not None
            if p.exitcode != 0:
                raise RuntimeError(f"Conversion failed with exit code {p.exitcode}!")

            # We evaluate the end2end efficiency so we count everything.
            cvt_time = time.time() - tstart
            if i == 0:
                config_file.write(f"{time_diffs.mean() + cvt_time},{to_name}\n")
            else:
                config_file.write(f"{time_diffs[i - 1] + cvt_time},{to_name}\n")
        except Exception as e:
            # Models generated by LEMON are sometimes not valid.
            # We evaluate the end2end efficiency so we count everything.
            cvt_time = time.time() - tstart
            if i == 0:
                config_file.write(f"{time_diffs.mean() + cvt_time},FAILURE\n")
            else:
                config_file.write(f"{time_diffs[i - 1] + cvt_time},FAILURE\n")
            print(e)
        config_file.flush()
        if time_diffs[:i].sum() + (time.time() - cvt_start_time) > args.tlimit:
            print(f"Time limit reached. {i+1} models converted.")
            break
    config_file.close()



import os
import sys
import random
import numpy as np
import pickle

from nnsmith.error import IncorrectResult
from nnsmith.backends import BackendFactory, mk_factory
from nnsmith.difftest import assert_allclose
from nnsmith.util import gen_one_input
from nnsmith.fuzz import simple_bug_report


def mcov_write(path):
    if path:
        with open(path + ".pkl", "wb") as f:
            pickle.dump(omax_fac._coverage_install().get_hitmap(), f)


def verify(backend_name, oracle_name, predicted, oracle=None):
    try:
        if oracle is not None:
            assert_allclose(predicted, oracle, backend_name, oracle_name)
    except IncorrectResult as e:
        return e
    return None


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--models", nargs="+", required=True, help="List to ONNX model paths"
    )
    parser.add_argument(
        "--backend", type=str, help="One of ort, trt, tvm, and xla", required=True
    )
    parser.add_argument("--device", type=str, default="cpu")
    parser.add_argument(
        "--memcov", type=str, default=None, help="Path to store memcov."
    )
    parser.add_argument(
        "--seed", type=int, default=233, help="to generate random input data"
    )
    parser.add_argument("--fuzz_max_nodes", type=int, help="parameter from fuzzer")
    parser.add_argument("--fuzz_seed", type=int, help="seed parameter from fuzzer")
    parser.add_argument("--fuzz_report_folder", type=str, help="parameter from fuzzer")
    parser.add_argument(
        "--clean_after_eval", action="store_true", help="rm models/oracle after eval"
    )
    # add fuzz_timeout?
    args = parser.parse_args()

    if args.fuzz_report_folder is None:
        print(
            "[WARNING] Bug report is not enabled as fuzzer parameters are not provided.",
            file=sys.stderr,
        )

    # Set global seed
    random.seed(args.seed)
    np.random.seed(args.seed)

    omax_fac = mk_factory(args.backend, args.device, optmax=True)
    omin_fac = mk_factory(args.backend, args.device, optmax=False)

    if args.memcov:
        assert omax_fac._coverage_install().get_now() is not None, "Memcov unavailable!"

    for i, path in enumerate(args.models):
        print(f"-> {path}", flush=True, file=sys.stderr)
        onnx_model = BackendFactory.get_onnx_proto(path)
        oracle_path = path.replace(".onnx", ".pkl")

        num_must_valid = False
        if os.path.exists(oracle_path):
            with open(oracle_path, "rb") as f:
                res = pickle.load(f)
                eval_inputs = res[0]
                torch_outputs = res[1]
                if len(res) == 3:
                    num_must_valid = res[2]
        else:
            print(f"No oracle found for model `{path}`", file=sys.stderr)
            input_spec, onames = BackendFactory.analyze_onnx_io(onnx_model)
            eval_inputs = gen_one_input(input_spec, 1, 2)
            torch_outputs = None  # No oracle.

        to_repro = f"python nnsmith/graph_gen.py --max_nodes {args.fuzz_max_nodes} --seed {args.fuzz_seed} --viz_graph"

        # Convenience function to report bugs.
        def report_bug(category, exception):
            simple_bug_report(
                report_folder=args.fuzz_report_folder,
                buggy_onnx_path=path,
                oracle_path=oracle_path,
                message=to_repro + "\n" + str(exception),
                bug_type=category + type(exception).__name__,
                backend=args.backend,
            )

        try:
            omax_exec = omax_fac.make_backend(onnx_model)
            pred_omax = omax_exec(eval_inputs)
            # We compare results iff NaN/Inf-free as we only consider confident bugs.
            if num_must_valid and all(
                np.isfinite(v).all() for v in torch_outputs.values()
            ):
                # FIXME: We assume the oracle is generated by PyTorch (as `fuzz.py`).
                err_tch_omax = verify(
                    args.backend + "-omax", "PyTorch", pred_omax, torch_outputs
                )
                if err_tch_omax:
                    if omin_fac is None:
                        # For TensorRT, there's no O0 option.
                        report_bug("torch-omax-", err_tch_omax)
                    else:
                        # RESULT INCONSISTENCY. But need to see if we can narrow it down to an optimzation bug.
                        omin_exec = omin_fac.make_backend(onnx_model)
                        err_omax_omin = verify(
                            args.backend + "-omax",
                            args.backend + "-omin",
                            pred_omax,
                            omin_exec(eval_inputs),
                        )
                        if err_omax_omin:
                            # We can confirm this is an optimization bug as O[opt-max] != O[opt-min]
                            report_bug("omin-omax-", err_omax_omin)
                        else:
                            # O[opt-max] == O[opt-min]: even opt-min shows the inconsistency.
                            # So the bug might be a fault of `PyTorch` or other parts in the compiler... Human assistant is needed.
                            report_bug("torch-omax-", err_tch_omax)
        except Exception as e:
            # Well, some error occured (compiler internal err | nnsmith impl error).
            report_bug("cie-", e)  # Compiler internal error.

        # remove after the model is tested. useful for locating the crashed model in the batch.
        if args.clean_after_eval:
            if os.path.exists(path):
                os.unlink(path)
            if os.path.exists(oracle_path):
                os.unlink(oracle_path)

        mcov_write(args.memcov)

    mcov_write(args.memcov)



"""Evaluation steps:
1. Get onnx models:
    - LEMON:
        Run LEMON to generate models (https://github.com/ganler/LEMON);
        python experiments/lemon_tf2onnx.py --lemon_output_dir /.../LEMON/lemon_outputs/ --onnx_dir ...
    - NNSMITH: TBD
    - GRAPH-FUZZ: TBD
2. Get source-level coverage: https://github.com/ganler/tvm/tree/coverage
    python experiments/cov_eval.py --model_dir ONNX_DIR --report_folder REPORT_FOLDER
"""

import os
import sys
import random
from time import time
import argparse
import subprocess

from tqdm import tqdm
import numpy as np

from nnsmith.util import mkdir

# CMD EXAMPLE:
# python experiments/cov_eval.py --model_dir lemon --report_folder test-cov --backend tvm --lib ../tvm/build/libtvm.so --llvm-version 14
# python experiments/cov_eval.py --model_dir test-onnx --report_folder test-cov-q --backend ort \
#  --lib '../onnxruntime/build/Linux/RelWithDebInfo/libonnxruntime_providers_shared.so ../onnxruntime/build/Linux/RelWithDebInfo/libonnxruntime.so'  --llvm-version 14

# Coverage lcov:          {i}.lcov
# Timing / # model:       stats.csv


def batched(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model_dir", type=str, required=True, help="Folder to the onnx models."
    )
    parser.add_argument("--report_folder", type=str, required=True)
    parser.add_argument(
        "--backend", type=str, default="tvm", help="One of ort, trt, tvm, and xla"
    )
    parser.add_argument("--device", type=str, default="cpu")
    parser.add_argument("--dp", type=int, default=250, help="# data point you want.")
    parser.add_argument("--dev", type=str, default="cpu", help="cpu/gpu")
    parser.add_argument("--sum", action="store_true", help="Use summary.")
    parser.add_argument("--resume", action="store_true", help="Resume eval.")
    parser.add_argument(
        "--seed", type=int, default=233, help="to generate random input data"
    )
    parser.add_argument("--memcov", action="store_true", help="Use memcov.")
    parser.add_argument(
        "--lib", type=str, default=None, help="path to instrumented library"
    )
    parser.add_argument(
        "--max_time",
        type=int,
        default=60 * 60 * 4,
        help="max time in seconds for coverage evaluation",
    )
    parser.add_argument(
        "--llvm-version",
        type=str,
        default="",
        help="version of llvm during coverage stuff. must align w/ tvm.",
    )
    parser.add_argument("-y", action="store_true")
    parser.add_argument("--keep_raw", action="store_true")
    args = parser.parse_args()

    # Set global seed
    print(f"==> Setting global seed to {args.seed}")
    random.seed(args.seed)
    np.random.seed(args.seed)

    HAS_LZ4 = os.system("lz4 --version > /dev/null 2>&1") == 0
    if not HAS_LZ4 and not args.memcov:
        print(
            "==> lz4 not found. Storing lcov w/o compression is disk-killing. Please install lz4!"
        )
        exit(1)

    if args.lib:
        lib_expr = ""
        for lib in args.lib.split():
            assert os.path.exists(lib), f"{lib} does not exist!"
            lib_expr += f" -object {os.path.realpath(lib)} "
    else:
        assert (
            args.memcov
        ), "you need to provide either --lib for ast cov or --memcov for memcov."

    if args.memcov and args.dp < 1000:
        print(
            "==> For memcov, you need at least 1k data points as we lose cov when it crashes."
        )
        print(f"==> Setting data point from {args.dp} to 1000.")
        args.dp = 1000

    if not args.resume:
        mkdir(args.report_folder, yes=args.y)
        # FORMAT:
        #   batch time (gen time + eval time)
        #   seed
        #   # models
        config_file = open(os.path.join(args.report_folder, "stats.csv"), "w")
        stderr_file = open(os.path.join(args.report_folder, "stderr.txt"), "w")

        next_batch_since_hist = 0
        process_time_sum = 0  # sum of all btime
    else:
        print("==> Resuming eval...")
        assert os.path.exists(
            os.path.join(args.report_folder, "stats.csv")
        ), "stats.csv does not exist!"
        import pandas as pd

        btimes = (
            pd.read_csv(
                os.path.join(args.report_folder, "stats.csv"), usecols=[0], header=None
            )
            .to_numpy()
            .squeeze()
        )
        process_time_sum = btimes.sum()

        history = open(os.path.join(args.report_folder, "stats.csv"), "r").readlines()
        if len(history) > 0 and not history[-1]:
            history.pop()

        next_batch_since_hist = (
            int(history[-1].rstrip().split(",")[-1].split(".")[0]) + 1
        )

        config_file = open(os.path.join(args.report_folder, "stats.csv"), "a")
        stderr_file = open(os.path.join(args.report_folder, "stderr.txt"), "a")

    def record(btime, n_models, seed, cov_name):
        config_file.write(f"{btime},{seed},{n_models},{cov_name}\n")
        config_file.flush()

    lines = open(os.path.join(args.model_dir, "gentime.csv"), "r").readlines()
    batch_size = max(len(lines) // args.dp, 1)

    print(f"==> Setting batch size: {batch_size}")
    batch_list = list(batched(lines, n=batch_size))

    # sometimes stuff got crashed and we will attribute the cost to the later batch.
    lagged_time = 0
    lagged_n_model = 0

    btime_begin = int(process_time_sum)
    start_time = time()

    with tqdm(total=args.max_time) as pbar:
        pbar.update(process_time_sum)
        pbar.refresh()

        for i in range(next_batch_since_hist, len(batch_list)):
            if process_time_sum > args.max_time:
                print(f"==> Timeout!")
                break

            batch = batch_list[i]
            btime = 0
            sum_path = os.path.join(args.report_folder, f"{i}.txt")

            # for source cov;
            lcov_name = f"{i}.lcov"
            lcov_path = os.path.join(args.report_folder, lcov_name)

            # for memcov
            memcov_name = f"{i}.memcov"
            memcov_path = os.path.join(args.report_folder, memcov_name)

            seed = random.getrandbits(32)

            model_batch = []
            for line in batch:
                tstr, mstr = line.rstrip("\n").split(",")
                btime += float(tstr)  # Generation time
                if mstr != "FAILURE":
                    model_batch.append(os.path.join(args.model_dir, mstr))

            profraw_path = os.path.join(args.report_folder, f"{i}.profraw")

            # Execute batch evaluation
            copied_env = os.environ.copy()
            # Path to store llvm profile.
            copied_env["LLVM_PROFILE_FILE"] = str(profraw_path)

            tstart = time()  # <=== START
            arguments = [
                "python",
                "experiments/batch_eval.py",
                "--models",
                *model_batch,
                "--backend",
                args.backend,
                "--device",
                args.device,
                "--seed",
                str(seed),
                "--fuzz_report_folder",
                args.report_folder,
            ]
            if args.memcov:
                arguments += ["--memcov", memcov_path]
            p = subprocess.Popen(
                arguments,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=copied_env,
            )
            _, errs = p.communicate()
            errs = errs.decode()
            exit_code = p.returncode

            if exit_code != 0:
                print(f"==> Batch {i} process crashed!")

            # Write stderr
            stderr_file.write(f"iter {i}: =================> EXIT CODE {exit_code}\n")
            if errs:
                stderr_file.write(errs)
            stderr_file.flush()

            if "$ORT.SKIP$" in errs:  # all models are unsupported by ort. skip it.
                if os.path.exists(profraw_path) and not args.keep_raw:
                    os.remove(profraw_path)
                continue

            btime += time() - tstart  # <=== ENDING
            process_time_sum += btime

            # Wrap up this batch.

            # Get coverage report
            if not args.memcov:
                if os.path.exists(profraw_path):
                    llvm_profdata = "llvm-profdata"
                    llvm_cov = "llvm-cov"
                    if args.llvm_version and str(args.llvm_version).isnumeric():
                        llvm_profdata += f"-{args.llvm_version}"
                        llvm_cov += f"-{args.llvm_version}"

                    profdata_path = os.path.join(args.report_folder, f"{i}.profdata")
                    # summary might be useless as it does not consider prior runs.
                    if 0 != os.system(
                        f"{llvm_profdata} merge -sparse {profraw_path} -o {profdata_path}"
                    ) or 0 != os.system(
                        f"{llvm_cov} export -instr-profile={profdata_path} -format=lcov {lib_expr} > {lcov_path}"
                    ):
                        print(f"Getting coverage failed!!", file=sys.stderr)
                    else:  # clean temporary files
                        if args.sum:
                            os.system(
                                f"{llvm_cov} report -instr-profile={profdata_path} {lib_expr} > {sum_path}"
                            )
                        assert 0 == os.system(f"lz4 {lcov_path} {lcov_path}.lz4")
                        if not args.keep_raw:
                            os.remove(profraw_path)
                            os.remove(profdata_path)
                            os.remove(lcov_path)
                else:
                    print(f"{profraw_path} does not exist...", file=sys.stderr)

            if os.path.exists(lcov_path + ".lz4"):
                record(
                    btime + lagged_time,
                    len(model_batch) + lagged_n_model,
                    seed,
                    cov_name=lcov_name,
                )
                lagged_time = 0
                lagged_n_model = 0
            elif os.path.exists(memcov_path + ".pkl"):
                record(
                    btime + lagged_time,
                    len(model_batch) + lagged_n_model,
                    seed,
                    cov_name=memcov_name,
                )
                lagged_time = 0
                lagged_n_model = 0
            else:
                # Means no lcov or memcov due to some LLVM issues.
                lagged_time += btime
                lagged_n_model += len(model_batch)

            pbar.update(int(time() - start_time + btime_begin) - pbar.n)
            pbar.set_description(f"batch tasks: {i}/{len(batch_list)}")
            pbar.refresh()

        config_file.close()
        stderr_file.close()



import os
import re
import argparse

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import pandas as pd
import numpy as np


SMALL_SIZE = 10
MEDIUM_SIZE = 13
BIGGER_SIZE = 18

plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
plt.rc("axes", titlesize=MEDIUM_SIZE)  # fontsize of the axes title
plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("legend", fontsize=MEDIUM_SIZE)  # legend fontsize
plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title

plt.rcParams.update({"text.usetex": True})

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--root", type=str, nargs="+", required=True)
    parser.add_argument("--output", type=str, default="results")
    parser.add_argument(
        "--rm_slowest", action="store_true", help="Remove the slowest run"
    )
    args = parser.parse_args()

    REGEX_PATTERN = "(\d+)-model-(\d+)-node-exp"
    node_sizes = []

    # Plot data
    # X: Time
    # Y: Succ. Rate

    def sort_by_time(time, succ_rate):
        time = np.array(time)
        succ_rate = np.array(succ_rate)
        sort_idx = time.argsort()
        return time[sort_idx], succ_rate[sort_idx]

    sampling_res = []
    grad_res = []
    proxy_res = []

    for nsize_folder in args.root:
        res = re.match(REGEX_PATTERN, nsize_folder)
        n_model, n_nodes = res.groups()
        node_sizes.append(int(n_nodes))

        sampling_time = []
        sampling_succ_rate = []

        grad_time = []
        grad_succ_rate = []

        proxy_time = []
        proxy_succ_rate = []

        for f in os.listdir(nsize_folder):
            if f.endswith(".csv") and f != "model_info.csv":
                data = pd.read_csv(os.path.join(nsize_folder, f))

                # Do not count the slowest iteration (1st iter usually)
                # as initialization takes some time.
                last_idx = -2 if args.rm_slowest else -1

                idx = data["sampling-time"].to_numpy().argsort()[:last_idx]
                sampling_time.append(data["sampling-time"][idx].mean())
                sampling_succ_rate.append(data["sampling-succ"][idx].mean())

                idx = data["grad-time"].to_numpy().argsort()[:last_idx]
                grad_time.append(data["grad-time"][idx].mean())
                grad_succ_rate.append(data["grad-succ"][idx].mean())

                idx = data["proxy-time"].to_numpy().argsort()[:last_idx]
                proxy_time.append(data["proxy-time"][idx].mean())
                proxy_succ_rate.append(data["proxy-succ"][idx].mean())
            elif f == "model_info.csv":
                data = pd.read_csv(os.path.join(nsize_folder, f))
                gentime = data["gen_time"]
                print(f"{gentime.mean()=}, {gentime.min()=}, {gentime.max()=}")

        # sort succ rate by time
        sampling_res.append(sort_by_time(sampling_time, sampling_succ_rate))
        grad_res.append(sort_by_time(grad_time, grad_succ_rate))
        proxy_res.append(sort_by_time(proxy_time, proxy_succ_rate))

    fig, ax = plt.subplots(figsize=(8, 3.2), constrained_layout=True)

    colors = ["dodgerblue", "violet", "green"]  # ['b', 'r', 'g']
    markers = ["1", ".", "+"]
    markercolor = "k"
    markersize = 10
    markeredgewidth = 1.2
    lw = 1.5

    max_time = 0

    for i in range(3):
        c = colors[i]
        alpha = 1 - 0.36 * i

        sampling_time, sampling_succ_rate = sampling_res[i]
        grad_time, grad_succ_rate = grad_res[i]
        proxy_time, proxy_succ_rate = proxy_res[i]

        # self *= 1000: sec -> milli
        sampling_time *= 1000
        grad_time *= 1000
        proxy_time *= 1000

        ax.plot(
            proxy_time,
            proxy_succ_rate,
            marker=markers[0],
            markeredgecolor=markercolor,
            markersize=markersize,
            markeredgewidth=markeredgewidth,
            linestyle="--",
            color=c,
            lw=lw,
        )
        ax.plot(
            grad_time,
            grad_succ_rate,
            marker=markers[1],
            markeredgecolor=markercolor,
            markersize=markersize,
            markeredgewidth=markeredgewidth,
            linestyle="--",
            color=c,
            lw=lw,
        )
        ax.plot(
            sampling_time,
            sampling_succ_rate,
            marker=markers[2],
            markeredgecolor=markercolor,
            markersize=markersize,
            markeredgewidth=markeredgewidth,
            linestyle="--",
            color=c,
            lw=lw,
        )

        max_time = max(max_time, sampling_time.max())

    ax.grid(True, linestyle=":", linewidth=0.5, alpha=0.5)

    lines = ax.get_lines()
    legend1 = plt.legend(
        [lines[i] for i in [0, 1, 2]],
        ["Gradient (Proxy Deriv.)", "Gradient", "Sampling"],
        loc="upper right",
        title="Searching Method",
    )

    patches = []
    for i in range(3):
        patches.append(mpatches.Patch(color=colors[i], label=node_sizes[i]))
    legend2 = plt.legend(
        handles=patches,
        loc="center right",
        bbox_to_anchor=(1, 0.36),
        title="Model Size",
    )

    ax.add_artist(legend1)
    ax.add_artist(legend2)

    ax.set_yticks(np.arange(0.6, 1.1, 0.1))
    ax.set_ylim(0.6, 1.0)

    ax.set_xticks(np.arange(0, 31, 5))
    ax.set_xlim(0, max_time + 0.5)

    ax.set_xlabel("Avg. Searching Time (millisecond)", fontweight="bold")
    ax.set_ylabel("Success Rate", fontweight="bold")

    plt.savefig(os.path.join(args.output, f"input-search-merge.pdf"))
    plt.savefig(os.path.join(args.output, f"input-search-merge.png"))



from enum import Enum, unique

import numpy as np


@unique
class DType(Enum):
    float16 = "float16"
    float32 = "float32"
    float64 = "float64"
    uint8 = "uint8"  # Support quantized models.
    uint16 = "uint16"
    uint32 = "uint32"
    uint64 = "uint64"
    int8 = "int8"
    int16 = "int16"
    int32 = "int32"
    int64 = "int64"
    bool = "bool"
    complex64 = "complex64"
    complex128 = "complex128"

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        s = super().__str__()
        assert s.startswith("DType."), s
        return s[len("DType.") :]

    def short(self) -> str:
        return {
            DType.float16: "f16",
            DType.float32: "f32",
            DType.float64: "f64",
            DType.uint8: "u8",
            DType.uint16: "u16",
            DType.uint32: "u32",
            DType.uint64: "u64",
            DType.int8: "i8",
            DType.int16: "i16",
            DType.int32: "i32",
            DType.int64: "i64",
            DType.complex64: "c64",
            DType.complex128: "c128",
            DType.bool: "b",
        }[self]

    @staticmethod
    def is_float(dtype):  # Don't use string. Make it well-formed.
        assert isinstance(dtype, DType)
        return dtype in [DType.float32, DType.float64]

    @staticmethod
    def from_str(s):
        return {
            "f16": DType.float16,
            "f32": DType.float32,
            "f64": DType.float64,
            "u8": DType.uint8,
            "i8": DType.int8,
            "i32": DType.int32,
            "i64": DType.int64,
            "c64": DType.complex64,
            "c128": DType.complex128,
            "float16": DType.float16,
            "float32": DType.float32,
            "float64": DType.float64,
            "uint8": DType.uint8,
            "uint16": DType.uint16,
            "uint32": DType.uint32,
            "uint64": DType.uint64,
            "int8": DType.int8,
            "int16": DType.int16,
            "int32": DType.int32,
            "int64": DType.int64,
            "complex64": DType.complex64,
            "complex128": DType.complex128,
            "bool": DType.bool,
        }[s]

    def numpy(self):
        return {
            DType.float16: np.float16,
            DType.float32: np.float32,
            DType.float64: np.float64,
            DType.uint8: np.uint8,
            DType.uint8: np.uint8,
            DType.uint16: np.uint16,
            DType.uint32: np.uint32,
            DType.uint64: np.uint64,
            DType.int8: np.int8,
            DType.int16: np.int16,
            DType.int32: np.int32,
            DType.int64: np.int64,
            DType.complex64: np.complex64,
            DType.complex128: np.complex128,
            DType.bool: np.bool_,
        }[self]

    def torch(self) -> "torch.dtype":
        import torch

        return {
            DType.float16: torch.float16,
            DType.float32: torch.float32,
            DType.float64: torch.float64,
            DType.uint8: torch.uint8,
            # PyTorch does not support other unsigned int types: https://github.com/pytorch/pytorch/issues/58734
            DType.int8: torch.int8,
            DType.int16: torch.int16,
            DType.int32: torch.int32,
            DType.int64: torch.int64,
            DType.complex64: torch.complex64,
            DType.complex128: torch.complex128,
            DType.bool: torch.bool,
        }[self]

    @staticmethod
    def from_torch(dtype) -> "DType":
        import torch

        return {
            torch.float16: DType.float16,
            torch.float32: DType.float32,
            torch.float64: DType.float64,
            torch.uint8: DType.uint8,
            torch.int8: DType.int8,
            torch.int16: DType.int16,
            torch.int32: DType.int32,
            torch.int64: DType.int64,
            torch.complex64: DType.complex64,
            torch.complex128: DType.complex128,
            torch.bool: DType.bool,
        }[dtype]

    def tensorflow(self) -> "tf.Dtype":
        import tensorflow as tf

        return {
            DType.float16: tf.float16,
            DType.float32: tf.float32,
            DType.float64: tf.float64,
            DType.uint8: tf.uint8,
            DType.uint16: tf.uint16,
            DType.uint32: tf.uint32,
            DType.uint64: tf.uint64,
            DType.int8: tf.int8,
            DType.int16: tf.int16,
            DType.int32: tf.int32,
            DType.int64: tf.int64,
            DType.complex64: tf.complex64,
            DType.complex128: tf.complex128,
            DType.bool: tf.bool,
        }[self]

    @staticmethod
    def from_tensorflow(dtype) -> "DType":
        import tensorflow as tf

        return {
            tf.float16: DType.float16,
            tf.float32: DType.float32,
            tf.float64: DType.float64,
            tf.uint8: DType.uint8,
            tf.uint16: DType.uint16,
            tf.uint32: DType.uint32,
            tf.uint64: DType.uint64,
            tf.int8: DType.int8,
            tf.int16: DType.int16,
            tf.int32: DType.int32,
            tf.int64: DType.int64,
            tf.complex64: DType.complex64,
            tf.complex128: DType.complex128,
            tf.bool: DType.bool,
        }[dtype]

    def sizeof(self) -> int:
        return {
            DType.float16: 2,
            DType.float32: 4,
            DType.float64: 8,
            DType.uint8: 1,
            DType.uint16: 2,
            DType.uint32: 4,
            DType.uint64: 8,
            DType.int8: 1,
            DType.int16: 2,
            DType.int32: 4,
            DType.int64: 8,
            DType.complex64: 8,
            DType.complex128: 16,
            DType.bool: 1,  # Follow C/C++ convention.
        }[self]


# "DTYPE_GEN*" means data types used for symbolic generation.
# "DTYPE_GEN_ALL" is surely a subset of all types but it is
# used to conservatively to avoid unsupported data types while
# applying nnsmith to various frameworks.
DTYPE_GEN_FLOATS = [DType.float16, DType.float32, DType.float64]
DTYPE_GEN_INTS = [
    DType.int8,
    DType.int16,
    DType.int32,
    DType.int64,
    DType.uint8,
]
DTYPE_GEN_COMPLEX = [DType.complex64, DType.complex128]
DTYPE_GEN_ALL = DTYPE_GEN_FLOATS + DTYPE_GEN_INTS + DTYPE_GEN_COMPLEX
DTYPE_GEN_NON_BOOL = [dtype for dtype in DTYPE_GEN_ALL if dtype != DType.bool]



import GPUtil
import pytest

has_gpu = len(GPUtil.getGPUs()) > 0

from nnsmith.abstract.dtype import DType
from nnsmith.backends import BackendFactory
from nnsmith.graph_gen import model_gen
from nnsmith.materialize import Model, TestCase
from nnsmith.narrow_spec import auto_opconfig, auto_opset

TestCase.__test__ = False  # supress PyTest warning


@pytest.mark.skipif(not has_gpu, reason="Skipping TensorRT testing due to no GPU found")
def test_narrow_spec_cache_make_and_reload():
    factory = BackendFactory.init("tensorrt", target="cuda", optmax=True)
    ONNXModel = Model.init("onnx")
    opset_lhs = auto_opconfig(ONNXModel, factory)
    assert opset_lhs, "Should not be empty... Something must go wrong."
    opset_rhs = auto_opconfig(ONNXModel, factory)
    assert opset_lhs == opset_rhs

    # Assert types
    assert isinstance(opset_lhs["core.ReLU"].in_dtypes[0][0], DType)

    # Assert Dictionary Type Equality
    assert type(opset_lhs) == type(opset_rhs)
    assert type(opset_lhs["core.ReLU"]) == type(opset_rhs["core.ReLU"])
    assert type(opset_lhs["core.ReLU"].in_dtypes[0][0]) == type(
        opset_rhs["core.ReLU"].in_dtypes[0][0]
    )


@pytest.mark.skipif(not has_gpu, reason="Skipping TensorRT testing due to no GPU found")
def test_synthesized_onnx_model(tmp_path):
    d = tmp_path / "test_trt_onnx"
    d.mkdir()

    ONNXModel = Model.init("onnx")
    factory = BackendFactory.init("tensorrt", target="cuda", optmax=True)

    gen = model_gen(
        opset=auto_opset(ONNXModel, factory),
        seed=23132,
        max_nodes=1,
    )  # One op should not be easily wrong... I guess.

    model = ONNXModel.from_gir(gen.make_concrete())

    assert model.with_torch

    model.refine_weights()  # either random generated or gradient-based.
    oracle = model.make_oracle()

    testcase = TestCase(model, oracle)
    testcase.dump(root_folder=d)

    assert factory.verify_testcase(testcase) is None



import os

ONNX_EXTERNAL_DATA_DIR_SUFFIX = "-mlist"
NNSMITH_ORT_INTRA_OP_THREAD = int(os.getenv("NNSMITH_ORT_INTRA_OP_THREAD", 1))
NNSMITH_BUG_PATTERN_TOKEN = "${PATTERN}"


def onnx2external_data_dir(onnx_file):
    return onnx_file + ONNX_EXTERNAL_DATA_DIR_SUFFIX



import onnx
import onnxruntime as ort
from multipledispatch import dispatch

from nnsmith.backends import BackendFactory
from nnsmith.backends.factory import BackendCallable
from nnsmith.macro import NNSMITH_ORT_INTRA_OP_THREAD
from nnsmith.materialize.onnx import ONNXModel

OPT_LEVELS = [
    ort.GraphOptimizationLevel.ORT_DISABLE_ALL,
    ort.GraphOptimizationLevel.ORT_ENABLE_BASIC,
    ort.GraphOptimizationLevel.ORT_ENABLE_EXTENDED,
    ort.GraphOptimizationLevel.ORT_ENABLE_ALL,
]


class ORTFactory(BackendFactory):
    def __init__(self, target, optmax, **kwargs):
        """opt_level ranges from 0 to 3, stands for ORT_DISABLE_ALL, ORT_ENABLE_BASIC, ORT_ENABLE_EXTENDED and ORT_ENABLE_ALL.
        See https://onnxruntime.ai/docs/performance/graph-optimizations.html for detail"""
        super().__init__(target, optmax, **kwargs)
        self.opt_level = OPT_LEVELS[-1 if optmax else 0]

        if target == "cuda":
            self.providers = [
                "CUDAExecutionProvider",
                "CPUExecutionProvider",
            ]  # ordered by precedence
        elif target == "cpu":
            self.providers = ["CPUExecutionProvider"]
        else:
            raise ValueError(
                f"Unknown target `{target}`. Only `cpu` and `cuda` are supported."
            )

    @property
    def system_name(self) -> str:
        return "onnxruntime"

    @property
    def version(self) -> str:
        return ort.__version__

    @dispatch(ONNXModel)
    def make_backend(
        self,
        model: ONNXModel,
    ) -> BackendCallable:
        sess_options = ort.SessionOptions()
        sess_options.graph_optimization_level = self.opt_level
        # https://github.com/microsoft/onnxruntime/issues/8313
        sess_options.intra_op_num_threads = NNSMITH_ORT_INTRA_OP_THREAD

        sess = ort.InferenceSession(
            onnx._serialize(model.native_model),
            providers=self.providers,
            sess_options=sess_options,
        )
        out_names = list(model.output_like.keys())

        def closure(inputs):
            res = sess.run(out_names, inputs)
            return {n: r for n, r in zip(out_names, res)}

        return closure



import pytest

from nnsmith.abstract.dtype import DType
from nnsmith.abstract.op import *
from nnsmith.gir import *


@mark_abstract("test")
class FakeSwap(AbsOpBase):
    in_dtypes = [(i, i) for i in DTYPE_GEN_ALL]
    out_dtypes = [(i, i) for i in DTYPE_GEN_ALL]

    def __init__(self):
        super().__init__()
        self.inp_ranks = [rank_all(), rank_all()]
        self.out_ranks = [rank_all(), rank_all()]

    def type_transfer(self, input_shapes: List[AbsTensor]) -> List[AbsTensor]:
        return [input_shapes[-1], input_shapes[0]]

    def deduct_inp_ranks_and_dtype(
        self, out_abs_tensor: List[AbsTensor]
    ) -> List[Tuple[int, DType]]:
        return [
            (out_abs_tensor[1].ndims, out_abs_tensor[1].dtype),
            (out_abs_tensor[0].ndims, out_abs_tensor[0].dtype),
        ]


def test_inst_expr():
    ph = Placeholder(ttype=AbsTensor(shape=[2, 3, 3], dtype=DType.float32))
    iexpr = InstExpr(ph, [])
    assert iexpr.op == ph
    assert iexpr.args == []
    assert iexpr.n_input() == 0
    assert iexpr.n_output() == 1

    op = Add()
    iexpr = InstExpr(op, ["a", "b"])
    assert iexpr.op == op
    assert iexpr.args == ["a", "b"]
    assert iexpr.n_input() == 2
    assert iexpr.n_output() == 1


def test_inst_ir():
    ph = Placeholder(ttype=AbsTensor(shape=[2, 3, 3], dtype=DType.float32))
    iexpr = InstExpr(ph, [])
    iir0 = InstIR(iexpr, irctx=[])
    assert iir0.iexpr == iexpr
    assert iir0.n_input() == 0
    assert iir0.n_output() == 1
    assert iir0.retval() == "v0_0"
    assert list(iir0.retvals()) == ["v0_0"]
    assert list(iir0.leaf_var()) == ["v0_0"]

    op = FakeSwap()
    iexpr = InstExpr(op, [iir0.retval(), iir0.retval()])
    iir1 = InstIR(iexpr, irctx=[iir0])
    assert iir1.iexpr == iexpr
    assert iir1.n_input() == 2
    assert iir1.n_output() == 2
    assert iir1.retval() == "v1_0"
    assert iir1.retval(0) == "v1_0"
    assert iir1.retval(1) == "v1_1"
    assert iir1.retvals() == ["v1_0", "v1_1"]

    inst_id, index = InstIR.var_inst_idx(iir1.retval(1))
    assert inst_id == 1  # The id of 2nd inst should be 1 (given `irctx`).
    assert index == 1

    # Check udchain
    assert iir1.is_user_of(iir0)  # This check actually does not rely on obj id.
    assert iir1 in iir0.users[0]
    assert iir1.no_users()
    assert not iir0.is_user_of(iir1, 0)
    assert not iir0.is_user_of(iir1, 1)

    # expect exception
    with pytest.raises(ValueError):
        iir1.is_user_of(iir0, 1)  # domain \in [0, 1)


def test_gir_mutate():
    ir = GraphIR()

    ph1 = ir.add_inst(
        InstExpr(Placeholder(ttype=AbsTensor(shape=[2, 3, 3], dtype=DType.float32)), [])
    )
    # Insert
    swap1 = ir.add_inst(InstExpr(FakeSwap(), [ph1.retval(), ph1.retval()]))

    assert ir.n_var() == 3
    assert ir.n_compute_inst() == 1
    assert len(ir.leaf_inst()) == 1
    assert ir.leaf_var() == [f"v{swap1.identifier}_0", f"v{swap1.identifier}_1"]

    assert (
        ir.pretty()
        == """v0_0 = Placeholder() \t# inst id: 0
v1_0, v1_1 = test.FakeSwap(v0_0, v0_0) \t# inst id: 1
"""
    )

    ir.assert_wellform()

    # IR: (ph1, ph1)=>(swap)
    # Backward insert a placeholder
    ph2 = ir.add_inst(
        InstExpr(Placeholder(ttype=AbsTensor(shape=[2, 3, 3], dtype=DType.float32)), [])
    )
    # IR: (ph2, ph1)=>(swap)
    ir.replace_arg(swap1, 0, ph2.retval())

    assert ir.n_var() == 4
    assert ir.n_compute_inst() == 1
    assert len(ir.leaf_inst()) == 1
    assert ir.leaf_var() == [f"v{swap1.identifier}_0", f"v{swap1.identifier}_1"]
    ir.assert_wellform()

    assert (
        ir.pretty()
        == """v0_0 = Placeholder() \t# inst id: 0
v1_0 = Placeholder() \t# inst id: 1
v2_0, v2_1 = test.FakeSwap(v0_0, v1_0) \t# inst id: 2
"""
    )

    # IR: (ph2, ph1)=>(swap1)=>(swap2)
    #                       \=>swap.1
    swap2 = ir.add_inst(InstExpr(FakeSwap(), [swap1.retval(0), swap1.retval(0)]))
    assert ir.n_var() == 6
    assert ir.n_compute_inst() == 2
    assert len(ir.leaf_inst()) == 1
    assert ir.leaf_var() == [
        f"v{swap1.identifier}_1",
        f"v{swap2.identifier}_0",
        f"v{swap2.identifier}_1",
    ]
    ir.assert_wellform()

    assert (
        ir.pretty()
        == """v0_0 = Placeholder() \t# inst id: 0
v1_0 = Placeholder() \t# inst id: 1
v2_0, v2_1 = test.FakeSwap(v0_0, v1_0) \t# inst id: 2
v3_0, v3_1 = test.FakeSwap(v2_0, v2_0) \t# inst id: 3
"""
    )

    # replace swap1 with add
    # IR: (ph2, ph1)=>(add)=>(swap2)
    add = ir.add_inst(InstExpr(Add(), [ph2.retval(), ph1.retval()]))
    ir.replace_alluse(swap1.retval(0), add.retval())
    ir.remove_unused(swap1)

    assert ir.n_var() == 5
    assert ir.n_compute_inst() == 2
    assert len(ir.leaf_inst()) == 1
    assert ir.leaf_var() == [f"v{swap2.identifier}_0", f"v{swap2.identifier}_1"]
    ir.assert_wellform()

    assert (
        ir.pretty()
        == """v0_0 = Placeholder() \t# inst id: 0
v1_0 = Placeholder() \t# inst id: 1
v2_0 = core.Add(v0_0, v1_0) \t# inst id: 2
v3_0, v3_1 = test.FakeSwap(v2_0, v2_0) \t# inst id: 3
"""
    )


def test_gir_repair():
    # Given a graph of wrong topological order or use-def chain, we can repair it.
    ir = GraphIR()
    ph = ir.add_inst(
        InstExpr(Placeholder(ttype=AbsTensor(shape=[2, 3, 3], dtype=DType.float32)), [])
    )
    swap = ir.add_inst(InstExpr(FakeSwap(), [ph.retval(), ph.retval()]))

    ir.assert_wellform()
    assert ph.users[0] == {swap}

    # 1.1 Repair missed use.
    ph.users[0] = set()  # violate the user
    with pytest.raises(AssertionError):
        ir.assert_wellform()

    ir._udchain_repair()
    ir.assert_wellform()
    assert ph.users[0] == {swap}

    # 1.2 Repair invalid use.
    swap.users[1] = {swap}  # create non-existing use
    with pytest.raises(AssertionError):
        ir.assert_wellform()

    ir._udchain_repair()
    ir.assert_wellform()
    assert not swap.users[0]
    assert not swap.users[1]

    # 2. Repair bad topological order.
    ir.insts[0], ir.insts[1] = ir.insts[1], ir.insts[0]
    with pytest.raises(AssertionError):
        ir.assert_wellform()

    ir._topological_sort()
    ir.assert_wellform()
    assert ir.insts[0] == ph
    assert ir.insts[1] == swap


def test_gir_dot():
    ir = GraphIR()

    ph1 = ir.add_inst(
        InstExpr(Placeholder(ttype=AbsTensor(shape=[2, 3, 3], dtype=DType.float32)), [])
    )
    ir.add_inst(InstExpr(FakeSwap(), [ph1.retval(), ph1.retval()]))

    assert (
        ir.pretty()
        == """v0_0 = Placeholder() \t# inst id: 0
v1_0, v1_1 = test.FakeSwap(v0_0, v0_0) \t# inst id: 1
"""
    )

    assert (
        ir.to_dot()
        == r"""digraph D {
  node [shape=Mrecord];
  0 [label="{Placeholder|{<o0> v0_0}}",fillcolor=lightgray,style=filled,];
  1 [label="{{<i0> v0_0|<i1> v0_0}|test.FakeSwap|{<o0> v1_0|<o1> v1_1}}",];
  0:o0 -> 1:i0 [label="f32[2, 3, 3]"];
  0:o0 -> 1:i1 [label="f32[2, 3, 3]"];
}
"""
    )
